# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy

import torch

import theseus as th


class MockVar(th.Manifold):
    def __init__(self, length, tensor=None, name=None):
        super().__init__(length, tensor=tensor, name=name)

    @staticmethod
    def _init_tensor(length):
        return torch.empty(1, length)

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        return True

    @staticmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    def dof(self):
        return 0

    def _local_impl(self, variable2):
        pass

    def _local_jacobian(self, var2):
        pass

    def _retract_impl(delta):
        pass

    def _copy_impl(self, new_name=None):
        return MockVar(self.tensor.shape[1], tensor=self.tensor.clone(), name=new_name)

    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        return euclidean_grad.clone()


class MockCostWeight(th.CostWeight):
    def __init__(
        self, the_data, name=None, add_dummy_var_with_name=None, add_optim_var=None
    ):
        super().__init__(name=name)
        if isinstance(the_data, torch.Tensor):
            the_data = th.Variable(the_data)
        self.the_data = the_data
        self.register_aux_var("the_data")
        if add_dummy_var_with_name:
            var = MockVar(1, name=add_dummy_var_with_name)
            setattr(self, add_dummy_var_with_name, var)
            self.register_optim_var(add_dummy_var_with_name)
        if add_optim_var:
            setattr(self, add_optim_var.name, add_optim_var)
            self.register_optim_var(add_optim_var.name)

    def weight_error(self, error):
        return self.the_data.tensor * error

    def weight_jacobians_and_error(self, jacobians, error):
        raise NotImplementedError(
            "weight_jacobians_and_error is not implemented for MockCostWeight."
        )

    def _copy_impl(self, new_name=None):
        return MockCostWeight(self.the_data.copy(), name=new_name)


class NullCostWeight(th.CostWeight):
    def __init__(self):
        super().__init__(name="null_cost_weight")

    def _init_tensor(self):
        pass

    def weight_error(self, error):
        return error

    def weight_jacobians_and_error(self, jacobians, error):
        return jacobians, error

    def _copy_impl(self, new_name=None):
        return NullCostWeight()


class MockCostFunction(th.CostFunction):
    def __init__(
        self, optim_vars, aux_vars, cost_weight, name=None, no_copy_vars=False
    ):
        super().__init__(cost_weight, name=name)
        for i, var in enumerate(optim_vars):
            attr_name = f"optim_var_{i}"
            setattr(self, attr_name, var)
            self.register_optim_var(attr_name)
        for i, aux in enumerate(aux_vars):
            attr_name = f"aux_var_{i}"
            setattr(self, attr_name, aux)
            self.register_aux_var(attr_name)
        self._dim = 2
        self._no_copy_vars = no_copy_vars

    def error(self):
        mu = torch.stack([v.tensor for v in self.optim_vars]).sum()
        return mu * torch.ones(1, self._dim)

    def jacobians(self):
        return [torch.ones(1, self._dim, self._dim)] * len(self._optim_vars_attr_names)

    def dim(self) -> int:
        return self._dim

    def _copy_impl(self, new_name=None):
        def _copy_fn(var):
            if self._no_copy_vars:
                return var
            else:
                return var.copy()

        return MockCostFunction(
            [_copy_fn(v) for v in self.optim_vars],
            [_copy_fn(aux) for aux in self.aux_vars],
            self.weight.copy(),
            name=new_name,
        )


def create_mock_cost_functions(tensor=None, cost_weight=NullCostWeight()):
    len_data = 1 if tensor is None else tensor.shape[1]
    var1 = MockVar(len_data, tensor=tensor, name="var1")
    var2 = MockVar(len_data, tensor=tensor, name="var2")
    var3 = MockVar(len_data, tensor=tensor, name="var3")
    aux1 = MockVar(len_data, tensor=tensor, name="aux1")
    aux2 = MockVar(len_data, tensor=tensor, name="aux2")
    names = [
        "MockCostFunction.var1.var2",
        "MockCostFunction.var1.var3",
        "MockCostFunction.var2.var3",
    ]
    cost_function_1_2 = MockCostFunction(
        [var1, var2], [aux1, aux2], cost_weight, name=names[0]
    )
    cost_function_1_3 = MockCostFunction(
        [var1, var3], [aux1], cost_weight, name=names[1]
    )
    cost_function_2_3 = MockCostFunction(
        [var2, var3], [aux2], cost_weight, name=names[2]
    )

    var_to_cost_functions = {
        var1: [cost_function_1_2, cost_function_1_3],
        var2: [cost_function_1_2, cost_function_2_3],
        var3: [cost_function_2_3, cost_function_1_3],
    }
    aux_to_cost_functions = {
        aux1: [cost_function_1_2, cost_function_1_3],
        aux2: [cost_function_1_2, cost_function_2_3],
    }
    return (
        [cost_function_1_2, cost_function_1_3, cost_function_2_3],
        names,
        var_to_cost_functions,
        aux_to_cost_functions,
    )


def create_objective_with_mock_cost_functions(
    tensor=None, cost_weight=NullCostWeight()
):
    (
        cost_functions,
        names,
        var_to_cost_functions,
        aux_to_cost_functions,
    ) = create_mock_cost_functions(tensor=tensor, cost_weight=cost_weight)

    objective = th.Objective()
    for cost_function in cost_functions:
        objective.add(cost_function)

    return (
        objective,
        cost_functions,
        names,
        var_to_cost_functions,
        aux_to_cost_functions,
    )


def check_copy_var(var: th.Variable):
    var.name = "old"
    new_var = var.copy(new_name="new")
    assert var is not new_var
    assert var.tensor is not new_var.tensor
    assert torch.allclose(var.tensor, new_var.tensor)
    assert new_var.name == "new"
    new_var_no_name = copy.deepcopy(var)
    assert new_var_no_name.name == f"{var.name}_copy"


def check_another_theseus_tensor_is_copy(var, other_var):
    assert isinstance(var, other_var.__class__)
    assert var is not other_var
    check_another_torch_tensor_is_copy(var.tensor, other_var.tensor)


def check_another_torch_tensor_is_copy(tensor, other_tensor):
    assert tensor is not other_tensor
    assert torch.allclose(tensor, other_tensor)


def check_another_theseus_function_is_copy(fn, other_fn, new_name):
    assert fn is not other_fn
    assert other_fn.name == new_name
    for var, new_var in zip(fn.optim_vars, other_fn.optim_vars):
        check_another_theseus_tensor_is_copy(var, new_var)
    for aux, new_aux in zip(fn.aux_vars, other_fn.aux_vars):
        check_another_theseus_tensor_is_copy(aux, new_aux)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Any, Dict, Optional, Type

import torch

from theseus.core import Objective
from theseus.optimizer import Linearization


class LinearSolver(abc.ABC):
    # linearization_cls is optional, since every linear solver will have a default
    def __init__(
        self,
        objective: Objective,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        linearization_kwargs = linearization_kwargs or {}
        self.linearization: Linearization = linearization_cls(
            objective, **linearization_kwargs
        )

    @abc.abstractmethod
    def solve(self, damping: Optional[float] = None, **kwargs) -> torch.Tensor:
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

from .vector import Vector


def _prepare_dof_and_tensor(
    expected_dof: int, tensor: Optional[torch.Tensor]
) -> Tuple[Optional[int], Optional[torch.Tensor]]:
    dof = None
    if tensor is None:
        dof = expected_dof
    else:
        if tensor.ndim == 1:
            tensor = tensor.view(1, -1)
        if tensor.shape[1] != expected_dof:
            raise ValueError(
                f"Provided tensor must have shape (batch_size, {expected_dof})."
            )
    return dof, tensor


class Point2(Vector):
    def __init__(
        self,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        dof, tensor = _prepare_dof_and_tensor(2, tensor)
        super().__init__(dof=dof, tensor=tensor, name=name, dtype=dtype)

    @staticmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "Point2":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        return Point2(
            tensor=torch.rand(
                size[0],
                2,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    @staticmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "Point2":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        return Point2(
            tensor=torch.randn(
                size[0],
                2,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        if tensor.ndim != 2 or tensor.shape[1] != 2:
            raise ValueError("Point2D data tensors can only be 2D vectors.")

        return True

    def __add__(self, other: Vector) -> "Point2":
        return cast(Point2, super().__add__(other))

    def __sub__(self, other: Vector) -> "Point2":
        return cast(Point2, super().__sub__(other))

    def __mul__(self, other: Union["Vector", torch.Tensor]) -> "Point2":
        return cast(Point2, super().__mul__(other))

    def __truediv__(self, other: Union["Vector", torch.Tensor]) -> "Point2":
        return cast(Point2, super().__truediv__(other))

    def __neg__(self) -> "Point2":
        return cast(Point2, super().__neg__())

    def cat(self, vecs: Union["Vector", Tuple["Vector"], List["Vector"]]) -> "Point2":
        return cast(Point2, super().cat(vecs))

    def abs(self) -> "Point2":
        return cast(Point2, super().abs())

    def x(self) -> torch.Tensor:
        return self[:, 0]

    def y(self) -> torch.Tensor:
        return self[:, 1]

    @staticmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "Point2":
        if tangent_vector.ndim != 2 or tangent_vector.shape[1] != 2:
            raise ValueError("Tangent vectors of Point2 should be 2-D vectors.")

        Vector._exp_map_jacobian_impl(tangent_vector, jacobians)

        return Point2(tensor=tangent_vector.clone())

    # added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "Point2":
        return cast(Point2, super().copy(new_name=new_name))


class Point3(Vector):
    def __init__(
        self,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        dof, tensor = _prepare_dof_and_tensor(3, tensor)
        super().__init__(dof=dof, tensor=tensor, name=name, dtype=dtype)

    @staticmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "Point3":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        return Point3(
            tensor=torch.rand(
                size[0],
                3,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    @staticmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "Point3":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        return Point3(
            tensor=torch.randn(
                size[0],
                3,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        if tensor.ndim != 2 or tensor.shape[1] != 3:
            raise ValueError("Point3 data tensors can only be 3D vectors.")

        return True

    def __add__(self, other: Vector) -> "Point3":
        return cast(Point3, super().__add__(other))

    def __sub__(self, other: Vector) -> "Point3":
        return cast(Point3, super().__sub__(other))

    def __mul__(self, other: Union["Vector", torch.Tensor]) -> "Point3":
        return cast(Point3, super().__mul__(other))

    def __truediv__(self, other: Union["Vector", torch.Tensor]) -> "Point3":
        return cast(Point3, super().__truediv__(other))

    def __neg__(self) -> "Point3":
        return cast(Point3, super().__neg__())

    def cat(self, vecs: Union["Vector", Tuple["Vector"], List["Vector"]]) -> "Point3":
        return cast(Point3, super().cat(vecs))

    def abs(self) -> "Point3":
        return cast(Point3, super().abs())

    def x(self) -> torch.Tensor:
        return self[:, 0]

    def y(self) -> torch.Tensor:
        return self[:, 1]

    def z(self) -> torch.Tensor:
        return self[:, 2]

    @staticmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "Point3":
        if tangent_vector.ndim != 2 or tangent_vector.shape[1] != 3:
            raise ValueError("Tangent vectors of Point3 should be 3-D vectors.")

        Vector._exp_map_jacobian_impl(tangent_vector, jacobians)

        return Point3(tensor=tangent_vector.clone())

    # added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "Point3":
        return cast(Point3, super().copy(new_name=new_name))


rand_point2 = Point2.rand
randn_point2 = Point2.randn
rand_point3 = Point3.rand
randn_point3 = Point3.randn



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .dense_linearization import DenseLinearization
from .linearization import Linearization
from .manifold_gaussian import ManifoldGaussian, local_gaussian, retract_gaussian
from .optimizer import Optimizer, OptimizerInfo
from .sparse_linearization import SparseLinearization
from .variable_ordering import VariableOrdering



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numdifftools as nd
import pytest  # noqa: F401
import torch

import theseus as th

torch.manual_seed(0)


def generate_data(num_points=10, a=1.0, b=0.5, noise_factor=0.01):
    data_x = torch.rand((1, num_points))
    noise = torch.randn((1, num_points)) * noise_factor
    data_y = a * data_x.square() + b + noise
    return data_x, data_y


num_points = 10
data_x, data_y = generate_data(num_points)

x = th.Variable(data_x.requires_grad_(), name="x")
y = th.Variable(data_y.requires_grad_(), name="y")
a = th.Vector(1, name="a")
b = th.Vector(1, name="b")


def quad_error_fn(optim_vars, aux_vars):
    a, b = optim_vars
    x, y = aux_vars
    est = a.tensor * x.tensor.square() + b.tensor
    err = y.tensor - est
    return err


optim_vars = [a, b]
aux_vars = [x, y]
cost_function = th.AutoDiffCostFunction(
    optim_vars,  # type: ignore
    quad_error_fn,
    num_points,
    aux_vars=aux_vars,
    name="quadratic_cost_fn",
)
objective = th.Objective()
objective.add(cost_function)
optimizer = th.GaussNewton(
    objective,
    max_iterations=15,
    step_size=1.0,
)

theseus_inputs = {
    "a": 2 * torch.ones((1, 1)).requires_grad_(),
    "b": torch.ones((1, 1)).requires_grad_(),
    "x": data_x,
    "y": data_y,
}
theseus_optim = th.TheseusLayer(optimizer)


def test_backwards():
    # First we use numdifftools to numerically compute the gradient
    # the optimal a w.r.t. the x part of the data
    def fit_x(data_x_np):
        theseus_inputs["x"] = (
            torch.from_numpy(data_x_np).float().clone().requires_grad_().unsqueeze(0)
        )
        updated_inputs, _ = theseus_optim.forward(
            theseus_inputs,
            optimizer_kwargs={"track_best_solution": True, "verbose": False},
        )
        return updated_inputs["a"].item()

    data_x_np = data_x.detach().clone().numpy()
    dfit_x = nd.Gradient(fit_x)
    da_dx_numeric = torch.from_numpy(dfit_x(data_x_np)).float()

    theseus_inputs["x"] = data_x
    updated_inputs, _ = theseus_optim.forward(
        theseus_inputs,
        optimizer_kwargs={
            "track_best_solution": True,
            "verbose": False,
            "backward_mode": th.BackwardMode.FULL,
        },
    )
    da_dx_full = torch.autograd.grad(updated_inputs["a"], data_x, retain_graph=True)[
        0
    ].squeeze()
    assert torch.allclose(da_dx_numeric, da_dx_full, atol=1e-3)

    updated_inputs, _ = theseus_optim.forward(
        theseus_inputs,
        optimizer_kwargs={
            "track_best_solution": True,
            "verbose": False,
            "backward_mode": th.BackwardMode.IMPLICIT,
        },
    )
    da_dx_implicit = torch.autograd.grad(
        updated_inputs["a"], data_x, retain_graph=True
    )[0].squeeze()
    assert torch.allclose(da_dx_numeric, da_dx_implicit, atol=1e-4)

    updated_inputs, _ = theseus_optim.forward(
        theseus_inputs,
        optimizer_kwargs={
            "track_best_solution": True,
            "verbose": False,
            "backward_mode": th.BackwardMode.TRUNCATED,
            "backward_num_iterations": 5,
        },
    )
    da_dx_truncated = torch.autograd.grad(
        updated_inputs["a"], data_x, retain_graph=True
    )[0].squeeze()
    assert torch.allclose(da_dx_numeric, da_dx_truncated, atol=1e-4)

    updated_inputs, _ = theseus_optim.forward(
        theseus_inputs,
        optimizer_kwargs={
            "track_best_solution": True,
            "verbose": False,
            "backward_mode": th.BackwardMode.DLM,
            "dlm_epsilon": 0.001,
        },
    )
    da_dx_truncated = torch.autograd.grad(
        updated_inputs["a"], data_x, retain_graph=True
    )[0].squeeze()
    assert torch.allclose(da_dx_numeric, da_dx_truncated, atol=1e-3)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, cast

import numpy as np
import torch
import torch.nn as nn

import theseus as th

from .motion_planner import MotionPlanner


class _ScalarModel(nn.Module):
    def __init__(self, hidden_size: int):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(1, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, 1),
        )

    def forward(self):
        dummy = torch.ones(1, 1).to(self.layers[0].weight.device)
        return self.layers(dummy)


class _OrderOfMagnitudeModel(nn.Module):
    def __init__(self, hidden_size: int, max_order: int):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(1, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, max_order),
            nn.ReLU(),
        )
        self.register_buffer("magnitudes", (10 ** torch.arange(max_order)).unsqueeze(0))

    def forward(self):
        dummy = torch.ones(1, 1).to(self.layers[0].weight.device)
        mag_weights = self.layers(dummy).softmax(dim=1)
        return (mag_weights * self.magnitudes).sum(dim=1, keepdim=True)


# ------------------------------------------------------------------------------------ #
# All public models in this module receive a batch of map data and return one or more
# torch tensors.
# ------------------------------------------------------------------------------------ #


class ScalarCollisionWeightModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.model = _OrderOfMagnitudeModel(10, 5)

    def forward(self, _: Dict[str, torch.Tensor]):
        return {"collision_w": self.model()}


class ScalarCollisionWeightAndCostEpstModel(nn.Module):
    def __init__(self, robot_radius: float):
        super().__init__()
        self.collision_weight_model = _OrderOfMagnitudeModel(200, 5)
        self.safety_dist_model = _ScalarModel(100)
        self.robot_radius = robot_radius

    def forward(self, _: Dict[str, torch.Tensor]):
        collision_w = self.collision_weight_model()
        safety_dist = self.safety_dist_model().sigmoid()
        return {"collision_w": collision_w, "cost_eps": safety_dist + self.robot_radius}


# High level idea: the trajectory is generated by the following sequence of steps:
#
# 1. Generate a parabola whose focus is the middle point between start and goal,
#    and whose distance between focus and vertex is the output of an MLP. Thus,
#    this part of the model learns how much it should bend outward from the straight
#    line path (and towards which side).
#
# 2. Add more precise curvatures to the trajectory by constructing a "sample" around
#    the parabola, using the Gaussian Process formulation of the original motion
#    planning problem (but ignoring obstacles). This relies on the fact that sampling
#    from the GP is equivalent to:
#
#          sample = Cholesky_Low(inverse(GP_covariance)) @ u,
#
#    where u is a multivariate vector sampled from a standard multivariate Normal
#    in the space of the GP variables.
#
#    Thus, we can generate arbitrary samples by learning the vector u to use, since
#    each value of the vector will be associated to a different trajectory. We learn the
#    vector u with another MLP, and use it to compute a fake "sample" around the
#    mean trajectory specified by the parabola.
#
#    Note: This basic model is included here for illustration purposes, and it
#    uses the map's file ID as input, so it has no ability to generalize. A more
#    practical extension would use a CNN encoder upstream to produce an input for
#    each individual map image.
class InitialTrajectoryModel(nn.Module):
    def __init__(
        self,
        planner: MotionPlanner,
        max_num_images: int = 1000,
        hid_size: int = 200,
    ):
        super().__init__()

        # This planner is only used to generate initial trajectories from a
        # multivariate gaussian around a curved trajectory,
        # and it's not part of the motion planning optimization that generates
        # the final trajectory
        self.aux_motion_planner = planner.copy(collision_weight=0.0)

        # Learns the vector u to use for generating the trajectory "sample", as
        # explained in Step 2 above.
        self.layers_u = nn.Sequential(
            nn.Linear(2 * max_num_images, hid_size),
            nn.ReLU(),
            nn.Linear(hid_size, hid_size),
            nn.ReLU(),
            nn.Linear(hid_size, 4 * (planner.num_time_steps + 1)),
        )

        # Learns a quadratic offset in normal direction to bend the mean trajectory.
        # This is the distance/direction between the parabola's focus and vertex,
        # mentioned in the Step 1 above.
        self.bend_factor = nn.Sequential(
            nn.Linear(2 * max_num_images, hid_size),
            nn.ReLU(),
            nn.Linear(hid_size, 1),
            nn.Tanh(),
        )

        def init_weights(m_):
            if isinstance(m_, nn.Linear):
                torch.nn.init.normal_(m_.weight)
                torch.nn.init.normal_(m_.bias)

        self.bend_factor.apply(init_weights)

        self.dt = planner.total_time / planner.num_time_steps

        self.num_images = max_num_images

    def forward(self, batch: Dict[str, Any]):
        device = self.aux_motion_planner.objective.device
        start = batch["expert_trajectory"][:, :2, 0].to(device)
        goal = batch["expert_trajectory"][:, :2, -1].to(device)

        one_hot_dummy = torch.zeros(start.shape[0], self.num_images * 2).to(device)
        file_ids = batch["file_id"]
        for batch_idx, fi in enumerate(file_ids):
            idx = int(fi.split("_")[1]) + int("forest" in fi) * self.num_images
            one_hot_dummy[batch_idx, idx] = 1

        # Compute straight line positions to use as mean of initial trajectory
        trajectory_len = self.aux_motion_planner.trajectory_len
        dist_vec = goal - start
        pos_incr_per_step = dist_vec / (trajectory_len - 1)
        trajectory = torch.zeros(start.shape[0], 4 * trajectory_len).to(
            device=device, dtype=start.dtype
        )
        trajectory[:, :2] = start
        for t_step in range(1, trajectory_len):
            idx = 4 * t_step
            trajectory[:, idx : idx + 2] = (
                trajectory[:, idx - 4 : idx - 2] + pos_incr_per_step
            )

        # Add the parabola bend
        bend_factor = self.bend_factor(one_hot_dummy)
        start_goal_dist = dist_vec.norm(dim=1)
        cur_t = torch.zeros_like(start_goal_dist) - start_goal_dist / 2
        c = (start_goal_dist / 2) ** 2
        angle = th.SO2(
            theta=torch.ones(dist_vec.shape[0], 1).to(device=device) * np.pi / 2
        )
        normal_vector = angle.rotate(th.Point2(tensor=dist_vec)).tensor.to(
            device=device
        )
        normal_vector /= normal_vector.norm(dim=1, keepdim=True)
        for t_step in range(1, trajectory_len):
            idx = 4 * t_step
            cur_t += start_goal_dist / (trajectory_len - 1)
            add = 2 * bend_factor * ((cur_t**2 - c) / c).view(-1, 1)
            trajectory[:, idx : idx + 2] += normal_vector * add

        # Compute resulting velocities
        for t_step in range(1, trajectory_len):
            idx = 4 * t_step
            trajectory[:, idx + 2 : idx + 4] = (
                trajectory[:, idx : idx + 2] - trajectory[:, idx - 4 : idx - 2]
            ) / self.dt

        # Finally, the final initial trajectory is further curved by a learned
        # "sample" from the GP describing the motion planning problem (ignoring
        # obstacles)
        # First, compute the covariance matrix
        with torch.no_grad():
            planner_inputs = {
                "sdf_origin": batch["sdf_origin"].to(device),
                "start": start.to(device),
                "goal": goal.to(device),
                "cell_size": batch["cell_size"].to(device),
                "sdf_data": batch["sdf_data"].to(device),
            }
            self.aux_motion_planner.objective.update(planner_inputs)

            motion_optimizer = cast(
                th.NonlinearLeastSquares, self.aux_motion_planner.layer.optimizer
            )
            linearization = cast(
                th.DenseLinearization, motion_optimizer.linear_solver.linearization
            )

            # Assign current trajectory as the mean of linearization
            for var in linearization.ordering:
                var_type, time_idx = var.name.split("_")
                assert var_type in ["pose", "vel"]
                if var_type == "pose":
                    traj_idx = int(time_idx) * 4
                if var_type == "vel":
                    traj_idx = int(time_idx) * 4 + 2
                var.update(trajectory[:, traj_idx : traj_idx + 2])

            linearization.linearize()
            cov_matrix = torch.inverse(linearization.AtA)
            lower_cov = torch.linalg.cholesky(cov_matrix)

        # Compute the u vector to generate the "sample"
        u = self.layers_u(one_hot_dummy).unsqueeze(2)
        initial_trajectory = trajectory.unsqueeze(2) + torch.matmul(lower_cov, u)

        # Construct the variable values dictionary to return
        values: Dict[str, torch.Tensor] = {}
        for t_step in range(trajectory_len):
            idx = 4 * t_step
            values[f"pose_{t_step}"] = initial_trajectory[:, idx : idx + 2, 0]
            values[f"vel_{t_step}"] = initial_trajectory[:, idx + 2 : idx + 4, 0]

        return values



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import logging
import pathlib
from typing import Any, Dict, List, Tuple

import omegaconf
import torch
import torch.nn as nn
import torch.optim as optim

import theseus as th

from .misc import TactilePushingDataset
from .models import (
    create_tactile_models,
    get_tactile_poses_from_values,
    update_tactile_pushing_inputs,
)
from .pose_estimator import TactilePoseEstimator

# Logger
logger = logging.getLogger(__name__)


# In this example, the goal is to track 2D object poses (x,y,theta) during planar
# pushing from tactile image measurements for single episode.
# This can solved as an optimization problem where the variables being estimated
# are object poses over time.
#
# We formulate the optimization using following cost terms:
#   * Quasi-static pushing planar: Penalizes deviation from quasi-static dynamics.
#     Uses velocity-only quasi-static model for sticking contact.
#   * Tactile measurements: Penalizes deviations from predicted relative pose
#     from tactile image feature pairs.
#   * Object-effector intersection: Penalizes intersections between object and
#     end-effector.
#   * End-effector priors: Penalizes deviation from end-effector poses captured from
#     motion capture.
#   * Boundary conditions: Penalizes deviation of first object pose from a global
#     pose prior.
#
# Based on the method described in,
# Sodhi et al. Learning Tactile Models for Factor Graph-based Estimation,
# 2021 (https://arxiv.org/abs/1705.10664)


class TactilePushingTrainer:
    def __init__(
        self, cfg: omegaconf.DictConfig, exp_path: pathlib.Path, device: torch.device
    ):
        self.cfg = cfg
        self.device = device
        dataset_path = exp_path / "datasets" / f"{cfg.dataset_name}.json"
        sdf_path = exp_path / "sdfs" / f"{cfg.sdf_name}.json"
        self.dataset_train = TactilePushingDataset(
            str(dataset_path),
            str(sdf_path),
            cfg.episode_length,
            cfg.train.batch_size,
            cfg.max_episodes,
            cfg.max_steps,
            device,
            split_episodes=cfg.split_episodes,
            data_mode="train",
        )
        self.dataset_val = TactilePushingDataset(
            str(dataset_path),
            str(sdf_path),
            cfg.episode_length,
            cfg.train.batch_size,
            cfg.max_episodes,
            cfg.max_steps,
            device,
            split_episodes=cfg.split_episodes,
            data_mode="val",
        )

        # -------------------------------------------------------------------- #
        # Create pose estimator (which wraps a TheseusLayer)
        # -------------------------------------------------------------------- #
        # self.dataset_train is used to created the correct SDF tensor shapes
        # and also get the number of time steps
        self.pose_estimator = TactilePoseEstimator(
            self.dataset_train,
            min_window_moving_frame=cfg.tactile_cost.min_win_mf,
            max_window_moving_frame=cfg.tactile_cost.max_win_mf,
            step_window_moving_frame=cfg.tactile_cost.step_win_mf,
            rectangle_shape=(cfg.shape.rect_len_x, cfg.shape.rect_len_y),
            device=device,
            optimizer_cls=getattr(th, cfg.inner_optim.optimizer),
            max_iterations=cfg.inner_optim.max_iters,
            step_size=cfg.inner_optim.step_size,
            regularization_w=cfg.inner_optim.reg_w,
            force_max_iters=cfg.inner_optim.force_max_iters,
        )

        # -------------------------------------------------------------------- #
        # Creating parameters to learn
        # -------------------------------------------------------------------- #
        if cfg.tactile_cost.init_pretrained_model:
            measurements_model_path = (
                exp_path / "models" / "transform_prediction_keypoints.pt"
            )
        else:
            measurements_model_path = None
        (
            self.measurements_model,
            self.qsp_model,
            self.mf_between_model,
            learnable_params,
        ) = create_tactile_models(
            cfg.train.mode, device, measurements_model_path=measurements_model_path
        )
        self.outer_optim = optim.Adam(learnable_params, lr=cfg.train.lr)

    def get_batch_data(
        self,
        batch: Dict[str, torch.Tensor],
        dataset: TactilePushingDataset,
        device: torch.device,
    ) -> Tuple[Dict[str, torch.Tensor], torch.Tensor, torch.Tensor]:
        # Initialize inputs dictionary
        theseus_inputs = self.pose_estimator.get_start_pose_and_motion_capture_dict(
            batch
        )
        # Update the above dictionary with measurement factor data
        update_tactile_pushing_inputs(
            dataset=dataset,
            batch=batch,
            measurements_model=self.measurements_model,
            qsp_model=self.qsp_model,
            mf_between_model=self.mf_between_model,
            device=device,
            cfg=self.cfg,
            theseus_inputs=theseus_inputs,
        )
        # Get ground truth data to use for the outer loss
        obj_poses_gt = batch["obj_poses_gt"]
        eff_poses_gt = batch["eff_poses_gt"]
        return theseus_inputs, obj_poses_gt, eff_poses_gt

    def _update(self, loss: torch.Tensor) -> Tuple[float, float]:
        start_event = torch.cuda.Event(enable_timing=True)
        end_event = torch.cuda.Event(enable_timing=True)
        start_event.record()
        torch.cuda.reset_max_memory_allocated()
        loss.backward()
        end_event.record()
        torch.cuda.synchronize()
        backward_time = start_event.elapsed_time(end_event)
        backward_mem = torch.cuda.max_memory_allocated() / 1025 / 1024
        logger.info(f"Backward pass took {backward_time} ms.")
        logger.info(f"Backward pass used {backward_mem} MBs.")

        nn.utils.clip_grad_norm_(self.qsp_model.parameters(), 100, norm_type=2)
        nn.utils.clip_grad_norm_(self.mf_between_model.parameters(), 100, norm_type=2)
        nn.utils.clip_grad_norm_(self.measurements_model.parameters(), 100, norm_type=2)

        with torch.no_grad():
            for name, param in self.qsp_model.named_parameters():
                logger.info(f"{name} {param.data}")
            for name, param in self.mf_between_model.named_parameters():
                logger.info(f"{name} {param.data}")

            def _print_grad(msg_, param_):
                logger.info(f"{msg_} {param_.grad.norm().item()}")

            _print_grad("    grad qsp", self.qsp_model.param)
            _print_grad("    grad mfb", self.mf_between_model.param)
            _print_grad("    grad nn_weight", self.measurements_model.fc1.weight)
            _print_grad("    grad nn_bias", self.measurements_model.fc1.bias)

        self.outer_optim.step()

        with torch.no_grad():
            for param in self.qsp_model.parameters():
                param.data.clamp_(0)
            for param in self.mf_between_model.parameters():
                param.data.clamp_(0)

        return backward_time, backward_mem

    def _resolve_backward_mode(self, epoch: int) -> th.BackwardMode:
        if epoch >= self.cfg.inner_optim.force_implicit_by_epoch - 1:  # 0-indexing
            logger.info("Forcing IMPLICIT backward mode.")
            return th.BackwardMode.IMPLICIT
        else:
            return getattr(th.BackwardMode, self.cfg.inner_optim.backward_mode)

    def compute_loss(
        self, epoch: int, update: bool = True
    ) -> Tuple[
        List[torch.Tensor], Dict[int, Dict[str, Any]], Dict[str, List[torch.Tensor]]
    ]:
        if update:
            dataset = self.dataset_train
        else:
            dataset = self.dataset_val

        results = {}
        losses = []
        image_data: Dict[str, List[torch.Tensor]] = dict(
            (name, []) for name in ["obj_opt", "eff_opt", "obj_gt", "eff_gt"]
        )

        # Set different number of max iterations for validation loop
        inner_opt_cfg = self.cfg.inner_optim
        self.pose_estimator.theseus_layer.optimizer.set_params(  # type: ignore
            max_iterations=inner_opt_cfg.max_iters
            if update or inner_opt_cfg.val_iters < 1
            else inner_opt_cfg.val_iters
        )
        for batch_idx in range(dataset.num_batches):
            # ---------- Read data from batch ----------- #
            batch = dataset.get_batch(batch_idx)
            theseus_inputs, obj_poses_gt, eff_poses_gt = self.get_batch_data(
                batch, dataset, self.device
            )

            # ---------- Forward pass ----------- #
            start_event = torch.cuda.Event(enable_timing=True)
            end_event = torch.cuda.Event(enable_timing=True)
            start_event.record()
            torch.cuda.reset_max_memory_allocated()
            theseus_outputs, info = self.pose_estimator.forward(
                theseus_inputs,
                optimizer_kwargs={
                    "verbose": True,
                    "track_err_history": True,
                    "backward_mode": self._resolve_backward_mode(epoch),
                    "backward_num_iterations": inner_opt_cfg.backward_num_iterations,
                    "dlm_epsilon": inner_opt_cfg.dlm_epsilon,
                    "__keep_final_step_size__": inner_opt_cfg.keep_step_size,
                },
            )
            end_event.record()
            torch.cuda.synchronize()
            forward_time = start_event.elapsed_time(end_event)
            forward_mem = torch.cuda.max_memory_allocated() / 1024 / 1024
            logger.info(f"Forward pass took {forward_time} ms.")
            logger.info(f"Forward pass used {forward_mem} MBs.")

            # ---------- Backward pass and update ----------- #
            obj_poses_opt, eff_poses_opt = get_tactile_poses_from_values(
                values=theseus_outputs, time_steps=dataset.time_steps
            )
            se2_opt = th.SE2(x_y_theta=obj_poses_opt.view(-1, 3))
            se2_gt = th.SE2(x_y_theta=obj_poses_gt.view(-1, 3))
            loss = se2_opt.local(se2_gt).norm()

            backward_time, backward_mem = -1.0, -1.0
            if update:
                backward_time, backward_mem = self._update(loss)

            # ---------- Pack results ----------- #
            losses.append(loss.item())
            results[batch_idx] = TactilePushingTrainer.pack_batch_results(
                theseus_outputs,
                self.qsp_model.state_dict(),
                self.mf_between_model.state_dict(),
                self.measurements_model.state_dict(),
                info,
                loss.item(),
                forward_time,
                backward_time,
                forward_mem,
                backward_mem,
            )
            image_data["obj_opt"].extend([p for p in obj_poses_opt])
            image_data["eff_opt"].extend([p for p in eff_poses_opt])
            image_data["obj_gt"].extend([p for p in obj_poses_gt])
            image_data["eff_gt"].extend([p for p in eff_poses_gt])

        return losses, results, image_data

    @staticmethod
    def pack_batch_results(
        theseus_outputs: Dict[str, torch.Tensor],
        qsp_state_dict: Dict[str, torch.Tensor],
        mfb_state_dict: Dict[str, torch.Tensor],
        meas_state_dict: Dict[str, torch.Tensor],
        info: th.optimizer.OptimizerInfo,
        loss_value: float,
        forward_time: float,
        backward_time: float,
        forward_mem: float,
        backward_mem: float,
    ) -> Dict[str, Any]:
        def _clone(t_):
            return t_.detach().cpu().clone()

        return {
            "theseus_outputs": dict((s, _clone(t)) for s, t in theseus_outputs.items()),
            "qsp_state_dict": qsp_state_dict,
            "mfb_state_dict": mfb_state_dict,
            "meas_state_dict": meas_state_dict,
            "err_history": info.err_history,  # type: ignore
            "loss": loss_value,
            "forward_time": forward_time,
            "backward_time": backward_time,
            "forward_mem": forward_mem,
            "backward_mem": backward_mem,
        }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest  # noqa: F401
import torch

import theseus as th

from .common import MockVar, check_copy_var


def test_copy():
    for i in range(100):
        size = torch.randint(low=1, high=21, size=(1,)).item()
        data = torch.rand(size=(1,) + (size,))
        var = MockVar(size, tensor=data, name="var")
        check_copy_var(var)


def test_var_shape():
    for sz in range(100):
        data = torch.ones(1, sz)
        var = MockVar(sz, tensor=data)
        assert data.shape == var.shape


def test_update_shape_check():
    for sz in range(2, 100):
        data = torch.ones(sz)
        var = MockVar(sz)
        with pytest.raises(ValueError):
            var.update(data)  # no batch dimension
        with pytest.raises(ValueError):
            var.update(data.view(-1, 1))  # wrong dimension
        var.update(data.view(1, -1))
        assert torch.isclose(var.tensor.squeeze(), data).all()


class MockVarNoArgs(th.Manifold):
    def __init__(self, tensor=None, name=None):
        super().__init__(tensor=tensor, name=name)

    @staticmethod
    def _init_tensor():
        return torch.ones(1, 1)

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        return True

    @staticmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    def dof(self):
        return 0

    def _local_impl(self, variable2):
        pass

    def _local_jacobian(self, var2):
        pass

    def _retract_impl(self, delta):
        pass

    def _copy_impl(self):
        return MockVarNoArgs()

    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        return euclidean_grad.clone()


def test_variable_no_args_init():
    var = MockVarNoArgs(name="mock")
    assert var.tensor.allclose(torch.ones(1, 1))
    assert var.name == "mock"
    var = MockVarNoArgs(tensor=torch.ones(2, 1))
    assert var.tensor.allclose(torch.ones(2, 1))
    var.update(torch.ones(3, 1))
    assert var.tensor.allclose(torch.ones(3, 1))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import warnings
from typing import Any, List, Optional

import torch

from theseus.constants import _CHECK_DTYPE_SUPPORTED
from theseus.core.variable import Variable

OptionalJacobians = Optional[List[torch.Tensor]]


# Abstract class to represent Manifold-type variables in the objective.
# Concrete classes must implement the following methods:
#   - `_init_tensor`: initializes the underlying tensor given constructor arguments.
#       The provided tensor must have a batch dimension.
#   -`_local`: given two close Manifolds gives distance in tangent space
#   - `_retract`: returns Manifold close by delta to given Manifold
#   - `update`: replaces the Manifold's tensor with the provided one (and checks that
#       the provided has the right format).
#   - `dof`: # of degrees of freedom of the variable
#
# Constructor can optionally provide an initial tensor value as a keyword argument.
class Manifold(Variable, abc.ABC):
    def __init__(
        self,
        *args: Any,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
        strict: bool = False,
    ):
        # If nothing specified, use torch's default dtype
        # else tensor.dtype takes precedence
        if tensor is None and dtype is None:
            dtype = torch.get_default_dtype()
        if tensor is not None:
            tensor = self._check_tensor(tensor, strict)
            if dtype is not None and tensor.dtype != dtype:
                warnings.warn(
                    f"tensor.dtype {tensor.dtype} does not match given dtype {dtype}, "
                    "tensor.dtype will take precendence."
                )
            dtype = tensor.dtype

        _CHECK_DTYPE_SUPPORTED(dtype)
        super().__init__(self.__class__._init_tensor(*args).to(dtype=dtype), name=name)
        if tensor is not None:
            self.update(tensor)

    # This method should return a tensor with the manifold's representation
    # as a function of the given args
    @staticmethod
    @abc.abstractmethod
    def _init_tensor(*args: Any) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def dof(self) -> int:
        pass

    @abc.abstractmethod
    def _local_impl(
        self, variable2: "Manifold", jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def _retract_impl(self, delta: torch.Tensor) -> "Manifold":
        pass

    @abc.abstractmethod
    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        pass

    def project(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        return self._project_impl(euclidean_grad, is_sparse)

    @staticmethod
    @abc.abstractmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        pass

    @staticmethod
    @abc.abstractmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        pass

    @classmethod
    def _check_tensor(cls, tensor: torch.Tensor, strict: bool = True) -> torch.Tensor:
        check = cls._check_tensor_impl(tensor)

        if not check:
            if strict:
                raise ValueError(f"The input tensor is not valid for {cls.__name__}.")
            else:
                tensor = cls.normalize(tensor)
                warnings.warn(
                    f"The input tensor is not valid for {cls.__name__} "
                    f"and has been normalized."
                )

        return tensor

    def local(
        self,
        variable2: "Manifold",
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> torch.Tensor:
        local_out = self._local_impl(variable2, jacobians)
        return local_out

    def retract(self, delta: torch.Tensor) -> "Manifold":
        return self._retract_impl(delta)

    # Must copy everything
    @abc.abstractmethod
    def _copy_impl(self, new_name: Optional[str] = None) -> "Manifold":
        pass

    def copy(self, new_name: Optional[str] = None) -> "Manifold":
        if not new_name:
            new_name = f"{self.name}_copy"
        var_copy = self._copy_impl(new_name=new_name)
        return var_copy

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        the_copy = self.copy()
        memo[id(self)] = the_copy
        return the_copy

    # calls to() on the internal tensors
    def to(self, *args, **kwargs):
        _, dtype, *_ = torch._C._nn._parse_to(*args, **kwargs)
        if dtype is not None:
            _CHECK_DTYPE_SUPPORTED(dtype)
        super().to(*args, **kwargs)


# Alias for Manifold.local()
def local(
    variable1: Manifold,
    variable2: Manifold,
    jacobians: Optional[List[torch.Tensor]] = None,
) -> torch.Tensor:
    return variable1.local(variable2, jacobians=jacobians)


# Alias for Manifold.retract()
def retract(variable: Manifold, delta: torch.Tensor) -> Manifold:
    return variable.retract(delta)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .dataset import (
    PoseGraphDataset,
    PoseGraphEdge,
    pg_histogram,
    read_2D_g2o_file,
    read_3D_g2o_file,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

import theseus as th


def _new_robust_cf(batch_size, loss_cls, generator) -> th.RobustCostFunction:
    v1 = th.rand_se3(batch_size, generator=generator)
    v2 = th.rand_se3(batch_size, generator=generator)
    w = th.ScaleCostWeight(torch.randn(1, generator=generator))
    cf = th.Local(v1, v2, w)
    ll_radius = th.Variable(tensor=torch.randn(1, 1, generator=generator))
    return th.RobustCostFunction(cf, loss_cls, ll_radius)


def _grad(jac: torch.Tensor, error: torch.Tensor) -> torch.Tensor:
    return jac.transpose(2, 1).matmul(error.unsqueeze(2)).permute(0, 2, 1)


def test_robust_cost_weighted_error():
    generator = torch.Generator()
    generator.manual_seed(0)
    for _ in range(10):
        for batch_size in [1, 2, 10]:
            for loss_cls in [th.WelschLoss, th.HuberLoss]:
                robust_cf = _new_robust_cf(batch_size, loss_cls, generator)
                cf = robust_cf.cost_function
                e = cf.weighted_error()
                rho = robust_cf.weighted_error()
                rho2 = (rho * rho).sum(dim=1, keepdim=True)
                # `RobustCostFunction.weighted_error` is written so that
                # ||we||2 == rho(||e||2)
                expected_rho2 = loss_cls.evaluate(
                    (e * e).sum(dim=1, keepdim=True), robust_cf.log_loss_radius.tensor
                )
                assert rho2.allclose(expected_rho2)


def test_robust_cost_grad_form():
    generator = torch.Generator()
    generator.manual_seed(0)
    for _ in range(10):
        for batch_size in [1, 2, 10]:
            for loss_cls in [th.WelschLoss, th.HuberLoss]:
                robust_cf = _new_robust_cf(batch_size, loss_cls, generator)
                cf = robust_cf.cost_function
                jacs, e = cf.weighted_jacobians_error()
                cf_grad = _grad(jacs[0], e)
                e_norm = (e * e).sum(1, keepdim=True)
                rho_prime = loss_cls.linearize(e_norm, robust_cf.log_loss_radius.tensor)
                # `weighted_jacobians_error()` is written so that it results in a
                # gradient equal to drho_de2 * J^T * e, which in the code is
                # `rho_prime * cf_grad`.
                expected_grad = rho_prime.view(-1, 1, 1) * cf_grad
                rescaled_jac, rescaled_e = robust_cf.weighted_jacobians_error()
                grad = _grad(rescaled_jac[0], rescaled_e)
                assert grad.allclose(expected_grad, atol=1e-6)


def test_robust_cost_jacobians():
    generator = torch.Generator()
    generator.manual_seed(0)

    for _ in range(10):
        for batch_size in [1, 2, 10]:
            for loss_cls in [th.WelschLoss, th.HuberLoss]:
                robust_cf = _new_robust_cf(batch_size, loss_cls, generator)
                v1, v2 = robust_cf.cost_function.var, robust_cf.cost_function.target
                v_aux = v1.copy()
                ll_radius = robust_cf.log_loss_radius
                w = robust_cf.cost_function.weight

                def test_fn(v_data):
                    v_aux.update(v_data)
                    new_robust_cf = th.RobustCostFunction(
                        th.Local(v_aux, v2, w), loss_cls, ll_radius
                    )
                    e = new_robust_cf.cost_function.weighted_error()
                    e_norm = (e * e).sum(1, keepdim=True)
                    return loss_cls.evaluate(e_norm, ll_radius.tensor) / 2.0

                aux_id = torch.arange(batch_size)
                grad_raw_dense = torch.autograd.functional.jacobian(
                    test_fn, (v1.tensor,)
                )[0]
                grad_raw_sparse = grad_raw_dense[aux_id, :, aux_id]
                expected_grad = v1.project(grad_raw_sparse, is_sparse=True)

                rescaled_jac, rescaled_err = robust_cf.weighted_jacobians_error()
                grad = _grad(rescaled_jac[0], rescaled_err)

                assert grad.allclose(expected_grad, atol=1e-2)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .gauss_newton import GaussNewton
from .levenberg_marquardt import LevenbergMarquardt
from .nonlinear_least_squares import NonlinearLeastSquares
from .nonlinear_optimizer import (
    BackwardMode,
    NonlinearOptimizer,
    NonlinearOptimizerParams,
    NonlinearOptimizerStatus,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import warnings

from .bundle_adjustment import BundleAdjustmentDataset, Camera, ba_histogram

try:
    from .motion_planning import (
        InitialTrajectoryModel,
        MotionPlanner,
        ScalarCollisionWeightAndCostEpstModel,
        ScalarCollisionWeightModel,
        TrajectoryDataset,
        generate_trajectory_figs,
    )
except ModuleNotFoundError:
    warnings.warn(
        "Unable to import Motion Planning utilities. "
        "Please make sure you have matplotlib installed."
    )

try:
    from .tactile_pose_estimation import (
        TactileMeasModel,
        TactilePoseEstimator,
        TactilePushingDataset,
        TactilePushingTrainer,
        TactileWeightModel,
        create_tactile_models,
        get_tactile_cost_weight_inputs,
        get_tactile_initial_optim_vars,
        get_tactile_motion_capture_inputs,
        get_tactile_nn_measurements_inputs,
        get_tactile_poses_from_values,
        init_tactile_model_from_file,
        update_tactile_pushing_inputs,
        visualize_tactile_push2d,
    )
except ModuleNotFoundError:
    warnings.warn(
        "Unable to import Tactile Pose Estimation utilities. "
        "Please make sure you have matplotlib and omegaconf installed."
    )


from .pose_graph import PoseGraphDataset, PoseGraphEdge, pg_histogram



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from theseus.constants import TEST_EPS
from theseus.utils import numeric_jacobian


def check_exp_map(tangent_vector, group_cls, atol=TEST_EPS):
    group = group_cls.exp_map(tangent_vector)
    tangent_vector_double = tangent_vector.double()
    tangent_vector_double.to(dtype=torch.float64)
    assert torch.allclose(
        group_cls.hat(tangent_vector_double).matrix_exp(),
        group.to_matrix().double(),
        atol=atol,
    )


def check_log_map(tangent_vector, group_cls, atol=TEST_EPS):
    assert torch.allclose(
        tangent_vector, group_cls.exp_map(tangent_vector).log_map(), atol=atol
    )


def check_compose(group_1, group_2):
    Jcmp = []
    composition = group_1.compose(group_2, jacobians=Jcmp)
    expected_matrix = group_1.to_matrix() @ group_2.to_matrix()
    group_1_double = group_1.copy()
    group_1_double.to(torch.float64)
    group_2_double = group_2.copy()
    group_2_double.to(torch.float64)
    expected_jacs = numeric_jacobian(
        lambda groups: groups[0].compose(groups[1]),
        [group_1_double, group_2_double],
    )
    batch = group_1.shape[0]
    dof = group_1.dof()
    assert torch.allclose(composition.to_matrix(), expected_matrix, atol=TEST_EPS)
    assert torch.allclose(Jcmp[0].double(), expected_jacs[0], atol=TEST_EPS)
    assert torch.allclose(
        Jcmp[1].double(),
        torch.eye(dof, dof, dtype=torch.float64).unsqueeze(0).expand(batch, dof, dof),
        atol=TEST_EPS,
    )
    if group_1.dtype == torch.float64:
        assert torch.allclose(Jcmp[1].double(), expected_jacs[1], atol=TEST_EPS)


def check_inverse(group):
    tangent_vector = group.log_map()
    inverse_group = group.exp_map(-tangent_vector.double())
    jac = []
    inverse_result = group.inverse(jacobian=jac)
    group_double = group.copy()
    group_double.to(torch.float64)
    expected_jac = numeric_jacobian(lambda groups: groups[0].inverse(), [group_double])
    assert torch.allclose(
        inverse_group.to_matrix().double(),
        inverse_result.to_matrix().double(),
        atol=TEST_EPS,
    )
    assert torch.allclose(jac[0].double(), expected_jac[0], atol=TEST_EPS)


def check_adjoint(group, tangent_vector):
    tangent_left = group.__class__.adjoint(group) @ tangent_vector.unsqueeze(2)
    group_matrix = group.to_matrix()
    tangent_right = group.__class__.vee(
        group_matrix.double()
        @ group.hat(tangent_vector.double())
        @ group.inverse().to_matrix().double()
    )
    assert torch.allclose(
        tangent_left.double().squeeze(2), tangent_right, atol=TEST_EPS
    )


# Func can be SO2.rotate, SE2.transform_to, SO3.unrotate, etc., whose third argument
# populates the jacobians
def check_projection_for_rotate_and_transform(
    Group, Point, Func, batch_size, generator=None, dtype=torch.float64
):
    group_double = Group.rand(batch_size, generator=generator, dtype=torch.float64)
    point_double = Point.rand(batch_size, generator=generator, dtype=torch.float64)
    group = group_double.copy()
    group.to(dtype)
    point = point_double.copy()
    point.to(dtype)

    def func(g, p):
        return Func(Group(tensor=g), p).tensor

    jac_raw = torch.autograd.functional.jacobian(
        func, (group_double.tensor, point_double.tensor)
    )

    if dtype == torch.float32:
        jac_raw = [jac_raw_n.float() for jac_raw_n in jac_raw]

    jac = []

    # Check returns
    rets = Func(group, point, jac)
    assert torch.allclose(
        rets.tensor.double(), func(group.tensor, point.tensor).double()
    )

    # Check dense jacobian matrices
    actual = [group.project(jac_raw[0]), point.project(jac_raw[1])]

    expected = [
        torch.zeros(batch_size, point.dof(), batch_size, group.dof()).double(),
        torch.zeros(batch_size, point.dof(), batch_size, point.dof()).double(),
    ]

    aux_id = torch.arange(batch_size)

    expected[0][aux_id, :, aux_id, :] = jac[0].double()
    expected[1][aux_id, :, aux_id, :] = jac[1].double()

    assert torch.allclose(actual[0].double(), expected[0], atol=TEST_EPS)
    assert torch.allclose(actual[1].double(), expected[1], atol=TEST_EPS)

    # Check sparse jacobian matrices
    actual = [
        group.project(jac_raw[0][aux_id, :, aux_id, :], is_sparse=True),
        point.project(jac_raw[1][aux_id, :, aux_id, :], is_sparse=True),
    ]

    expected = jac
    assert torch.allclose(actual[0], expected[0], atol=TEST_EPS)
    assert torch.allclose(actual[1], expected[1], atol=TEST_EPS)


def check_projection_for_compose(
    Group, batch_size, generator=None, dtype=torch.float64
):
    group1_double = Group.rand(batch_size, generator=generator, dtype=torch.float64)
    group2_double = Group.rand(batch_size, generator=generator, dtype=torch.float64)
    group1 = group1_double.copy()
    group1.to(dtype)
    group2 = group2_double.copy()
    group2.to(dtype)

    jac = []
    rets = group1.compose(group2, jacobians=jac)

    def func(g1, g2):
        return Group(tensor=g1).compose(Group(tensor=g2)).to_matrix()

    jac_raw = torch.autograd.functional.jacobian(
        func, (group1_double.tensor, group2_double.tensor)
    )

    # Check returns
    assert torch.allclose(rets.to_matrix(), func(group1.tensor, group2.tensor))

    # Check for dense jacobian matrices
    if dtype == torch.float32:
        jac_raw = [jac_raw_n.float() for jac_raw_n in jac_raw]

    temp = [group1.project(jac_raw[0]), group2.project(jac_raw[1])]
    actual = [
        torch.zeros(batch_size, rets.dof(), batch_size, rets.dof(), dtype=dtype),
        torch.zeros(batch_size, rets.dof(), batch_size, rets.dof(), dtype=dtype),
    ]

    for n in torch.arange(batch_size):
        for i in torch.arange(rets.dof()):
            actual[0][:, :, n, i] = rets.vee(
                rets.inverse().to_matrix() @ temp[0][:, :, :, n, i]
            )
            actual[1][:, :, n, i] = rets.vee(
                rets.inverse().to_matrix() @ temp[1][:, :, :, n, i]
            )

    expected = [
        torch.zeros(batch_size, rets.dof(), batch_size, rets.dof(), dtype=dtype),
        torch.zeros(batch_size, rets.dof(), batch_size, rets.dof(), dtype=dtype),
    ]

    aux_id = torch.arange(batch_size)

    expected[0][aux_id, :, aux_id, :] = jac[0]
    expected[1][aux_id, :, aux_id, :] = jac[1]

    assert torch.allclose(actual[0], expected[0], atol=TEST_EPS)
    assert torch.allclose(actual[1], expected[1], atol=TEST_EPS)

    # Check for sparse jacobian matrices
    temp = [
        group1.project(jac_raw[0][aux_id, :, :, aux_id], is_sparse=True),
        group2.project(jac_raw[1][aux_id, :, :, aux_id], is_sparse=True),
    ]
    actual = [
        torch.zeros(batch_size, rets.dof(), rets.dof(), dtype=dtype),
        torch.zeros(batch_size, rets.dof(), rets.dof(), dtype=dtype),
    ]

    for i in torch.arange(rets.dof()):
        actual[0][:, :, i] = rets.vee(rets.inverse().to_matrix() @ temp[0][:, :, :, i])
        actual[1][:, :, i] = rets.vee(rets.inverse().to_matrix() @ temp[1][:, :, :, i])

    expected = jac

    assert torch.allclose(actual[0], expected[0], atol=TEST_EPS)
    assert torch.allclose(actual[1], expected[1], atol=TEST_EPS)


def check_projection_for_inverse(
    Group, batch_size, generator=None, dtype=torch.float64
):
    group_double = Group.rand(batch_size, generator=generator, dtype=torch.float64)
    group = group_double.copy()
    group.to(dtype)

    jac = []
    rets = group.inverse(jacobian=jac)

    def func(g):
        return Group(tensor=g).inverse().to_matrix()

    jac_raw = torch.autograd.functional.jacobian(func, (group_double.tensor))

    if dtype == torch.float32:
        jac_raw = jac_raw.float()

    # Check returns
    assert torch.allclose(rets.to_matrix(), func(group.tensor), atol=TEST_EPS)

    # Check for dense jacobian matrices
    temp = group.project(jac_raw)
    actual = torch.zeros(batch_size, group.dof(), batch_size, group.dof(), dtype=dtype)

    for n in torch.arange(batch_size):
        for i in torch.arange(group.dof()):
            actual[:, :, n, i] = group.vee(group.to_matrix() @ temp[:, :, :, n, i])

    expected = torch.zeros(
        batch_size, group.dof(), batch_size, group.dof(), dtype=dtype
    )

    aux_id = torch.arange(batch_size)

    expected[aux_id, :, aux_id, :] = jac[0]

    assert torch.allclose(actual, expected, atol=TEST_EPS)

    # Check for sparse jacobian matrices
    temp = group.project(jac_raw[aux_id, :, :, aux_id], is_sparse=True)
    actual = torch.zeros(batch_size, group.dof(), group.dof(), dtype=dtype)

    for i in torch.arange(group.dof()):
        actual[:, :, i] = group.vee(group.to_matrix() @ temp[:, :, :, i])

    expected = jac[0]

    assert torch.allclose(actual, expected, atol=TEST_EPS)


def check_projection_for_exp_map(tangent_vector, Group, is_projected=True, atol=1e-8):
    batch_size = tangent_vector.shape[0]
    dof = tangent_vector.shape[1]
    aux_id = torch.arange(batch_size)

    def exp_func(xi):
        return Group.exp_map(xi).to_matrix()

    actual = []
    _ = Group.exp_map(tangent_vector, jacobians=actual).to_matrix()

    tangent_vector_double = tangent_vector.double()
    group_double = Group.exp_map(tangent_vector_double).to_matrix()
    jac_raw = torch.autograd.functional.jacobian(exp_func, (tangent_vector_double))

    if is_projected:
        jac_raw = jac_raw[aux_id, :, :, aux_id]
        expected = torch.cat(
            [
                Group.vee(group_double.inverse() @ jac_raw[:, :, :, i]).view(-1, dof, 1)
                for i in torch.arange(dof)
            ],
            dim=2,
        )
    else:
        expected = jac_raw[aux_id, :, aux_id]

    assert torch.allclose(actual[0].double(), expected, atol=atol)


def check_projection_for_log_map(tangent_vector, Group, is_projected=True, atol=1e-8):
    batch_size = tangent_vector.shape[0]
    aux_id = torch.arange(batch_size)
    group_double = Group.exp_map(tangent_vector.double())
    group = group_double.copy()
    if tangent_vector.dtype == torch.float32:
        group.tensor = group.tensor.float()

    def log_func(group):
        return Group(tensor=group).log_map()

    jac_raw = torch.autograd.functional.jacobian(log_func, (group_double.tensor))[
        aux_id, :, aux_id
    ]

    if is_projected:
        expected = group_double.project(jac_raw, is_sparse=True)
    else:
        expected = jac_raw

    actual = []
    _ = group.log_map(jacobians=actual)

    assert torch.allclose(actual[0].double(), expected, atol=atol)


def check_jacobian_for_local(group0, group1, Group, is_projected=True, atol=TEST_EPS):
    batch_size = group0.shape[0]
    aux_id = torch.arange(batch_size)

    group0_double = group0.copy()
    group0_double.to(torch.float64)
    group1_double = group1.copy()
    group1_double.to(torch.float64)

    def local_func(group0, group1):
        return Group(tensor=group0).local(Group(tensor=group1))

    jac_raw = torch.autograd.functional.jacobian(
        local_func, (group0_double.tensor, group1_double.tensor)
    )

    if is_projected:
        expected = [
            group0_double.project(jac_raw[0][aux_id, :, aux_id], is_sparse=True),
            group1_double.project(jac_raw[1][aux_id, :, aux_id], is_sparse=True),
        ]
    else:
        expected = [jac_raw[0][aux_id, :, aux_id], jac_raw[1][aux_id, :, aux_id]]

    actual = []
    _ = group0.local(group1, jacobians=actual)

    assert torch.allclose(actual[0].double(), expected[0], atol=atol)
    assert torch.allclose(actual[1].double(), expected[1], atol=atol)


def check_normalize(group, batch_size, dtype):
    rng = torch.Generator()
    rng.manual_seed(0)

    matrix = group.rand(batch_size, dtype=dtype, generator=rng).tensor
    group_mat = group.normalize(matrix)
    torch.allclose(group_mat, matrix)

    matrix = torch.rand(matrix.shape, dtype=dtype)
    group_mat = group.normalize(matrix)
    group._check_tensor(group_mat)


def check_so3_se3_normalize(group, batch_size, dtype):
    rng = torch.Generator()
    rng.manual_seed(0)

    check_normalize(group, batch_size, dtype)

    matrix = group.rand(batch_size, dtype=dtype).tensor
    matrix[:, :, 2] *= -1
    group_mat = group.normalize(matrix)
    torch.allclose(
        (group_mat - matrix).norm(dim=[1, 2]),
        2 * torch.ones(matrix.shape[0], dtype=dtype),
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch  # needed for import of Torch C++ extensions to work
from scipy.sparse import csr_matrix

from theseus.utils import random_sparse_binary_matrix


def check_mat_mult(batch_size, num_rows, num_cols, fill, verbose=False):
    if not torch.cuda.is_available():
        return
    from theseus.extlib.mat_mult import apply_damping, mat_vec, mult_MtM, tmat_vec

    A_skel = random_sparse_binary_matrix(
        num_rows, num_cols, fill, min_entries_per_col=3
    )
    A_num_cols = num_cols
    A_rowPtr = torch.tensor(A_skel.indptr, dtype=torch.int).cuda()
    A_colInd = torch.tensor(A_skel.indices, dtype=torch.int).cuda()
    A_num_rows = A_rowPtr.size(0) - 1
    A_nnz = A_colInd.size(0)
    A_val = torch.rand((batch_size, A_nnz), dtype=torch.double).cuda()

    A_csr = [
        csr_matrix(
            (A_val[i].cpu(), A_colInd.cpu(), A_rowPtr.cpu()), (A_num_rows, A_num_cols)
        )
        for i in range(batch_size)
    ]
    if verbose:
        print("A[0]:\n", A_csr[0].todense())

    # test At * A
    AtA_csr = [(a.T @ a).tocsr() for a in A_csr]
    AtA_rowPtr = torch.tensor(AtA_csr[0].indptr).cuda()
    AtA_colInd = torch.tensor(AtA_csr[0].indices).cuda()
    AtA_val = torch.tensor(np.array([m.data for m in AtA_csr])).cuda()
    AtA_num_rows = AtA_rowPtr.size(0) - 1
    AtA_num_cols = AtA_num_rows
    AtA_nnz = AtA_colInd.size(0)  # noqa: F841

    if verbose:
        print("\nAtA[0]:\n", AtA_csr[0].todense())

    res = mult_MtM(batch_size, A_rowPtr, A_colInd, A_val, AtA_rowPtr, AtA_colInd)
    if verbose:
        print(
            "res[0]:\n",
            csr_matrix(
                (res[0].cpu(), AtA_colInd.cpu(), AtA_rowPtr.cpu()),
                (AtA_num_rows, AtA_num_cols),
            ).todense(),
        )

    assert AtA_val.isclose(res, atol=1e-10).all()

    # test damping
    old_diagonals = torch.tensor(
        np.array(
            [
                csr_matrix(
                    (res[x].cpu(), AtA_colInd.cpu(), AtA_rowPtr.cpu()),
                    (AtA_num_rows, AtA_num_cols),
                ).diagonal()
                for x in range(batch_size)
            ]
        )
    )
    alpha = 0.3
    beta = 0.7
    apply_damping(batch_size, AtA_num_cols, AtA_rowPtr, AtA_colInd, res, alpha, beta)
    new_diagonals = torch.tensor(
        np.array(
            [
                csr_matrix(
                    (res[x].cpu(), AtA_colInd.cpu(), AtA_rowPtr.cpu()),
                    (AtA_num_rows, AtA_num_cols),
                ).diagonal()
                for x in range(batch_size)
            ]
        )
    )
    assert new_diagonals.isclose(old_diagonals * (1 + alpha) + beta, atol=1e-10).all()

    # test A * b
    v = torch.rand((batch_size, A_num_cols), dtype=torch.double).cuda()
    A_v = torch.tensor(
        np.array([A_csr[i] @ v[i].cpu() for i in range(batch_size)])
    ).cuda()

    A_v_test = mat_vec(batch_size, A_num_cols, A_rowPtr, A_colInd, A_val, v)

    if verbose:
        print("A_v:", A_v)
        print("A_v_test:", A_v_test)

    assert A_v.isclose(A_v_test, atol=1e-10).all()

    # test At * b
    w = torch.rand((batch_size, A_num_rows), dtype=torch.double).cuda()
    At_w = torch.tensor(
        np.array([A_csr[i].T @ w[i].cpu() for i in range(batch_size)])
    ).cuda()

    At_w_test = tmat_vec(batch_size, A_num_cols, A_rowPtr, A_colInd, A_val, w)

    if verbose:
        print("A_w:", At_w)
        print("A_w_test:", At_w_test)

    assert At_w.isclose(At_w_test, atol=1e-10).all()


@pytest.mark.cudaext
def test_mat_mult_1():
    check_mat_mult(batch_size=5, num_rows=50, num_cols=30, fill=0.2)


@pytest.mark.cudaext
def test_mat_mult_2():
    check_mat_mult(batch_size=5, num_rows=150, num_cols=60, fill=0.2)


@pytest.mark.cudaext
def test_mat_mult_3():
    check_mat_mult(batch_size=10, num_rows=300, num_cols=90, fill=0.2)


@pytest.mark.cudaext
def test_mat_mult_4():
    check_mat_mult(batch_size=5, num_rows=50, num_cols=30, fill=0.1)


@pytest.mark.cudaext
def test_mat_mult_5():
    check_mat_mult(batch_size=5, num_rows=150, num_cols=60, fill=0.1)


@pytest.mark.cudaext
def test_mat_mult_6():
    check_mat_mult(batch_size=10, num_rows=300, num_cols=90, fill=0.1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from sksparse.cholmod import Factor as CholeskyDecomposition

from ..linear_system import SparseStructure


class CholmodSolveFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, *args, **kwargs):
        At_val: torch.Tensor = args[0]
        b: torch.Tensor = args[1]
        sparse_structure: SparseStructure = args[2]
        symbolic_decomposition: CholeskyDecomposition = args[3]
        damping: float = args[4]

        At_val_cpu = At_val.cpu().double()
        b_cpu = b.cpu().double()
        batch_size = At_val.shape[0]
        targs = {"dtype": At_val.dtype, "device": "cpu"}
        x_cpu = torch.empty(size=(batch_size, sparse_structure.num_cols), **targs)
        cholesky_decompositions = []

        for i in range(batch_size):

            # compute decomposition from symbolic decomposition
            At_i = sparse_structure.csc_transpose(At_val_cpu[i, :])
            cholesky_decomposition = symbolic_decomposition.cholesky_AAt(At_i, damping)

            # solve
            Atb_i = At_i @ b_cpu[i, :]
            x_cpu[i, :] = torch.Tensor(cholesky_decomposition(Atb_i))
            cholesky_decompositions.append(cholesky_decomposition)

        ctx.b_cpu = b_cpu
        ctx.x_cpu = x_cpu
        ctx.At_val_cpu = At_val_cpu
        ctx.sparse_structure = sparse_structure
        ctx.cholesky_decompositions = cholesky_decompositions

        return x_cpu.to(device=At_val.device, dtype=At_val.dtype)

    # Let v row vector, and w column vector of dimension n, m, and
    # A an nxm matrix. Then
    #   v * A * w = Sum(v_i * A_ij * w_j) = [v (X) w] . A
    # Where by [v (X) w] we mean the nxm matrix which is the
    # tensor product of v and w, and "." is the componentwise
    # dot product of the two nxm matrices.
    #
    # Now, we have
    #      At * A * x = At * b
    #
    # Therefore if A, b, x are parametrized (eg. by u) we have deriving
    #
    # (i)  At'*A*x + At*A'*x + At*A*x' = At'*b + At*b'
    #
    # indicating A'=dA/du, b'=db/du, x'=dx/du
    #
    # Now, assume we have a function f of x, and G = df/dx be the
    # gradient that we consider a row vector, so G*x' is df/du.
    #
    # To compute df/db and df/dA, make x' explicit in the (i):
    #
    #      x' = (At * A)^{-1} (At*b' + At'*b - At'*A*x - At*A'*x)
    #
    # So multiplying by the row vector G we have
    #
    #   G*x' = G*(At*A)^{-1}*At * b' + G*(At*A)^{-1} * (At'*b - At'*A*x - At*A'*x)
    #        = H * At * b' + H * At' * (b - A*x) - H * At * A' * x
    #        = (H*At) * b' + [H (X) (b-A*x)] . At' - [H*At (X) x] . A'
    #        = (H*At) * b' + [(b-A*x) (X) H - H*At (X) x] . A'
    # after putting H = G*(At*A)^{-1} for convenience.
    #
    # Therefore after switching to column vectors we have
    #   df/db = A*H
    # (where H = (At*A)^{-1}*G), while
    #   df/dA = (b-A*x) (X) H - A*H (X) x
    # The two tensor products means that to compute the gradient of
    # a block of A we have to multiply entries taken from (b-A*x) and H,
    # and blocks taken from A*H and x.
    #
    # Here we assume we are provided x and H after the linear solver
    # has been applied to Atb and the gradient G.

    # NOTE: in the torch docs the backward is also marked as "staticmethod", I think it makes sense
    @staticmethod
    def backward(ctx, grad_output):

        batch_size = grad_output.shape[0]
        targs = {"dtype": grad_output.dtype, "device": "cpu"}  # grad_output.device}
        H = torch.empty(size=(batch_size, ctx.sparse_structure.num_cols), **targs)
        AH = torch.empty(size=(batch_size, ctx.sparse_structure.num_rows), **targs)
        b_Ax = ctx.b_cpu.clone()
        grad_output_cpu = grad_output.cpu()

        for i in range(batch_size):

            H[i, :] = torch.Tensor(
                ctx.cholesky_decompositions[i](grad_output_cpu[i, :])
            )

            A_i = ctx.sparse_structure.csr_straight(ctx.At_val_cpu[i, :])
            AH[i, :] = torch.Tensor(A_i @ H[i, :])
            b_Ax[i, :] -= torch.Tensor(A_i @ ctx.x_cpu[i, :])

        # now we fill values of a matrix with structure identical to A with
        # selected entries from the difference of tensor products:
        #   b_Ax (X) H - AH (X) x
        # NOTE: this row-wise manipulation can be much faster in C++ or Cython
        A_col_ind = ctx.sparse_structure.col_ind
        A_row_ptr = ctx.sparse_structure.row_ptr
        batch_size = grad_output.shape[0]
        A_grad = torch.empty(
            size=(batch_size, len(A_col_ind)),
            device="cpu",
        )  # return value, A's grad
        for r in range(len(A_row_ptr) - 1):
            start, end = A_row_ptr[r], A_row_ptr[r + 1]
            columns = A_col_ind[start:end]  # col indices, for this row
            A_grad[:, start:end] = (
                b_Ax[:, r].unsqueeze(1) * H[:, columns]
                - AH[:, r].unsqueeze(1) * ctx.x_cpu[:, columns]
            )

        dev = grad_output.device
        return A_grad.to(device=dev), AH.to(device=dev), None, None, None



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.core.cost_weight import ScaleCostWeight

from .common import (
    MockCostFunction,
    MockCostWeight,
    MockVar,
    check_another_theseus_function_is_copy,
    create_mock_cost_functions,
)


def test_copy():
    cost_weight = MockCostWeight(th.Variable(torch.ones(1)))
    data = torch.ones(1, 1)
    cost_functions, *_ = create_mock_cost_functions(
        tensor=data, cost_weight=cost_weight
    )

    for cost_function in cost_functions:
        cost_function.weight = cost_weight

        def _check_new_cost_function(new_cost_function):
            check_another_theseus_function_is_copy(
                cost_function, new_cost_function, new_name=f"{cost_function.name}_copy"
            )
            check_another_theseus_function_is_copy(
                cost_weight,
                new_cost_function.weight,
                new_name=f"{cost_weight.name}_copy",
            )

        _check_new_cost_function(cost_function.copy())
        _check_new_cost_function(copy.deepcopy(cost_function))


def test_default_name_and_ids():
    reps = 100
    seen_ids = set()
    for i in range(reps):
        cost_function = MockCostFunction([], [], MockCostWeight(torch.ones(1)))
        cost_function_name = f"MockCostFunction__{cost_function._id}"
        seen_ids.add(cost_function._id)
        assert cost_function.name == cost_function_name
    assert len(seen_ids) == reps


@pytest.mark.parametrize("autograd_loop_over_batch", [True, False])
def test_autodiff_cost_function_error_and_jacobians_shape(autograd_loop_over_batch):
    for i in range(100):
        num_optim_vars = np.random.randint(0, 5)
        num_aux_vars = np.random.randint(0, 5)
        batch_size = np.random.randint(1, 10)
        err_dim = np.random.randint(1, 5)
        optim_vars = []
        aux_vars = []
        variable_values = torch.randn(num_optim_vars + num_aux_vars)
        idx = 0
        for i in range(num_optim_vars):
            optim_vars.append(
                MockVar(
                    idx + 1,
                    tensor=torch.ones(batch_size, idx + 1) * variable_values[idx],
                    name=f"optim_var_{i}",
                )
            )
            idx += 1
        for i in range(num_aux_vars):
            aux_vars.append(
                MockVar(
                    idx + 1,
                    tensor=torch.ones(batch_size, idx + 1) * variable_values[idx],
                    name=f"aux_var_{i}",
                )
            )
            idx += 1
        cost_weight = MockCostWeight(torch.ones(1, 1))

        # checks that the right number of optimization variables is passed
        # checks that the variable values are correct
        # returns the sum of the first elements of each tensor, which should be the
        # same as the sum of variables_values
        def error_fn(optim_vars, aux_vars):
            assert isinstance(optim_vars, tuple)
            assert len(optim_vars) == num_optim_vars
            assert len(optim_vars) > 0
            assert len(aux_vars) == num_aux_vars
            ret_val = torch.zeros(optim_vars[0].shape[0], err_dim)

            all_vars = optim_vars + aux_vars

            vals = []
            for i, arg in enumerate(all_vars):
                assert isinstance(arg, th.Variable)
                assert arg.shape == (batch_size, i + 1) or arg.shape == (1, i + 1)
                assert arg.tensor.allclose(
                    variable_values[i] * torch.ones_like(arg.tensor)
                )
                vals.append(arg[0, 0])
            return ret_val + torch.Tensor(vals).sum()

        # this checks that 0 optimization variables is not allowed
        if len(optim_vars) < 1:
            with pytest.raises(ValueError):
                th.AutoDiffCostFunction(
                    optim_vars,
                    error_fn,
                    1,
                    cost_weight=cost_weight,
                    aux_vars=aux_vars,
                )
        else:
            # check that the error function returns the correct value
            cost_function = th.AutoDiffCostFunction(
                optim_vars,
                error_fn,
                err_dim,
                cost_weight=cost_weight,
                aux_vars=aux_vars,
                autograd_loop_over_batch=autograd_loop_over_batch,
            )
            err = cost_function.error()
            assert err.allclose(variable_values.sum() * torch.ones(batch_size, err_dim))

            # Now checking the jacobians
            jacobians, err_jac = cost_function.jacobians()
            assert err_jac.allclose(err)
            assert len(jacobians) == num_optim_vars
            for i in range(num_optim_vars):
                # variable dim is i + 1 (see MockVar creation line)
                assert jacobians[i].shape == (batch_size, err_dim, i + 1)


@pytest.mark.parametrize("autograd_loop_over_batch", [True, False])
def test_autodiff_cost_function_cost_weight(autograd_loop_over_batch):
    batch_size = 10
    optim_vars = []
    aux_vars = []

    for i in range(5):
        optim_vars.append(
            MockVar(
                1,
                tensor=torch.ones(batch_size, 1) * torch.randn(1),
                name=f"optim_var_{i}",
            )
        )
        aux_vars.append(
            MockVar(
                1,
                tensor=torch.ones(batch_size, 1) * torch.randn(1),
                name=f"aux_var_{i}",
            )
        )

    def error_fn(optim_vars, aux_vars):
        assert len(optim_vars) > 0
        return torch.ones(optim_vars[0].shape[0], 1)

    # test verifying default CostWeight
    cost_function = th.AutoDiffCostFunction(
        optim_vars,
        error_fn,
        1,
        aux_vars=aux_vars,
        autograd_loop_over_batch=autograd_loop_over_batch,
    )
    assert isinstance(cost_function.weight, ScaleCostWeight)
    assert torch.allclose(cost_function.weight.scale.tensor, torch.ones(1, 1))
    weighted_error = cost_function.weighted_error()
    assert torch.allclose(weighted_error, torch.ones(batch_size, 1))

    # test overriding default CostWeight
    for i in range(10):
        cost_weight_value = torch.randn(1, 1)
        cost_weight = MockCostWeight(cost_weight_value)
        cost_function = th.AutoDiffCostFunction(
            optim_vars,
            error_fn,
            1,
            cost_weight=cost_weight,
            aux_vars=aux_vars,
        )
        assert cost_function.weight is cost_weight
        assert torch.allclose(
            cost_function.weight.the_data.tensor, cost_weight_value
        )  # type: ignore
        weighted_error = cost_function.weighted_error()
        direct_error_computation = cost_weight_value * torch.ones(batch_size, 1)
        assert torch.allclose(weighted_error, direct_error_computation)


@pytest.mark.parametrize("autograd_loop_over_batch", [True, False])
def test_autodiff_cost_function_to(autograd_loop_over_batch):
    batch_size = 10
    optim_vars = []
    aux_vars = []

    for i in range(5):
        optim_vars.append(
            MockVar(
                1,
                tensor=torch.ones(batch_size, 1) * torch.randn(1),
                name=f"optim_var_{i}",
            )
        )
        aux_vars.append(
            MockVar(
                1,
                tensor=torch.ones(batch_size, 1) * torch.randn(1),
                name=f"aux_var_{i}",
            )
        )

    def error_fn(optim_vars, aux_vars):
        res = 0
        for var in optim_vars:
            res += var.tensor
        return res

    # test verifying default CostWeight
    cost_function = th.AutoDiffCostFunction(
        optim_vars,
        error_fn,
        1,
        aux_vars=aux_vars,
        autograd_loop_over_batch=autograd_loop_over_batch,
    )

    for var in optim_vars:
        var.to(dtype=torch.double)

    # This fails because internal vars of the cost function have not been converted
    # to double
    with pytest.raises(ValueError):
        cost_function.jacobians()

    cost_function.to(dtype=torch.double)
    cost_function.jacobians()


@pytest.mark.parametrize("autograd_loop_over_batch", [True, False])
def test_autodiff_cost_function_error_and_jacobians_shape_on_SO3(
    autograd_loop_over_batch,
):
    for i in range(100):
        num_vars = np.random.randint(0, 5)
        batch_size = np.random.randint(1, 10)
        err_dim = 3
        optim_vars = []
        aux_vars = []
        idx = 0
        for i in range(num_vars):
            optim_vars.append(th.SO3.rand(batch_size, dtype=torch.float64))
            idx += 1
        for i in range(num_vars):
            aux_vars.append(th.Point3.rand(batch_size, dtype=torch.float64))
            idx += 1
        cost_weight = MockCostWeight(torch.ones(1, 1))

        def error_fn(optim_vars, aux_vars):
            assert isinstance(optim_vars, tuple)
            assert len(optim_vars) == num_vars
            assert len(optim_vars) > 0
            assert len(aux_vars) == num_vars
            ret_val = torch.zeros(optim_vars[0].shape[0], err_dim)

            for optim_var, aux_var in zip(optim_vars, aux_vars):
                ret_val += th.SO3(tensor=optim_var.tensor).rotate(aux_var).tensor

            return ret_val

        # this checks that 0 optimization variables is not allowed
        if len(optim_vars) < 1:
            with pytest.raises(ValueError):
                th.AutoDiffCostFunction(
                    optim_vars,
                    error_fn,
                    1,
                    cost_weight=cost_weight,
                    aux_vars=aux_vars,
                )
        else:
            # check that the error function returns the correct value
            cost_function = th.AutoDiffCostFunction(
                optim_vars,
                error_fn,
                err_dim,
                cost_weight=cost_weight,
                aux_vars=aux_vars,
                autograd_loop_over_batch=autograd_loop_over_batch,
            )
            err = cost_function.error()

            # Now checking the jacobians
            jacobians, err_jac = cost_function.jacobians()
            assert err_jac.allclose(err)
            assert len(jacobians) == num_vars
            for i in range(num_vars):
                # variable dim is i + 1 (see MockVar creation line)
                assert jacobians[i].shape == (batch_size, err_dim, 3)


@pytest.mark.parametrize("autograd_loop_over_batch", [True, False])
def test_autodiff_cost_function_error_and_jacobians_value_on_SO3(
    autograd_loop_over_batch,
):
    for i in range(100):
        num_vars = np.random.randint(0, 5)
        batch_size = np.random.randint(1, 10)
        err_dim = 3
        optim_vars = []
        aux_vars = []
        idx = 0
        for i in range(num_vars):
            optim_vars.append(th.SO3.rand(batch_size, dtype=torch.float64))
            idx += 1
        for i in range(num_vars):
            aux_vars.append(th.Point3.rand(batch_size, dtype=torch.float64))
            idx += 1
        cost_weight = MockCostWeight(torch.ones(1, 1, dtype=torch.float64))

        def error_fn(optim_vars, aux_vars):
            assert isinstance(optim_vars, tuple)
            assert len(optim_vars) == num_vars
            assert len(optim_vars) > 0
            assert len(aux_vars) == num_vars
            ret_val = torch.zeros(optim_vars[0].shape[0], err_dim, dtype=torch.float64)

            for optim_var, aux_var in zip(optim_vars, aux_vars):
                ret_val += th.SO3(tensor=optim_var.tensor).rotate(aux_var).tensor

            return ret_val

        # this checks that 0 optimization variables is not allowed
        if len(optim_vars) < 1:
            with pytest.raises(ValueError):
                th.AutoDiffCostFunction(
                    optim_vars,
                    error_fn,
                    1,
                    cost_weight=cost_weight,
                    aux_vars=aux_vars,
                )
        else:
            # check that the error function returns the correct value
            cost_function = th.AutoDiffCostFunction(
                optim_vars,
                error_fn,
                err_dim,
                cost_weight=cost_weight,
                aux_vars=aux_vars,
                autograd_loop_over_batch=autograd_loop_over_batch,
            )
            jac_actual, err_actual = cost_function.jacobians()

            err_expected = torch.zeros(batch_size, 3, dtype=torch.float64)
            for n in torch.arange(num_vars):
                jac = []  # type: ignore
                err_expected += optim_vars[n].rotate(aux_vars[n], jacobians=jac).tensor
                assert torch.allclose(jac_actual[n], jac[0])

            assert torch.allclose(err_actual, err_expected)



from typing import Dict, List, Optional, Tuple, Type

import numpy as np
import torch

import theseus as th
from theseus.optimizer.nonlinear.levenberg_marquardt import LevenbergMarquardt

from .misc import TactilePushingDataset


class TactilePoseEstimator:
    def __init__(
        self,
        dataset: TactilePushingDataset,
        min_window_moving_frame: int,
        max_window_moving_frame: int,
        step_window_moving_frame: int,
        rectangle_shape: Tuple[float, float],
        device: torch.device,
        optimizer_cls: Optional[Type[th.NonlinearLeastSquares]] = LevenbergMarquardt,
        max_iterations: int = 3,
        step_size: float = 1.0,
        regularization_w: float = 0.0,
        force_max_iters: bool = False,
    ):
        time_steps = dataset.time_steps

        # -------------------------------------------------------------------- #
        # Creating optimization variables
        # -------------------------------------------------------------------- #
        # The optimization variables for this problem are SE2 object and end effector
        # poses over time.
        obj_poses, eff_poses = [], []
        for i in range(time_steps):
            obj_poses.append(th.SE2(name=f"obj_pose_{i}", dtype=torch.double))
            eff_poses.append(th.SE2(name=f"eff_pose_{i}", dtype=torch.double))

        # -------------------------------------------------------------------- #
        # Creating auxiliary variables
        # -------------------------------------------------------------------- #
        #  - obj_start_pose: target for boundary cost functions
        #  - motion_captures: priors on the end-effector poses
        #  - nn_measurements: tactile measurement prediction from image features
        #  - sdf_data, sdf_cell_size, sdf_origin: signed distance field data,
        #    cell_size and origin
        obj_start_pose = th.SE2(name="obj_start_pose")
        self.obj_start_pose = obj_start_pose

        motion_captures: List[th.SE2] = []
        for i in range(time_steps):
            motion_captures.append(th.SE2(name=f"motion_capture_{i}"))
        self.motion_captures = motion_captures

        nn_measurements = []
        for i in range(min_window_moving_frame, time_steps):
            for offset in range(
                min_window_moving_frame,
                np.minimum(i, max_window_moving_frame),
                step_window_moving_frame,
            ):
                nn_measurements.append(th.SE2(name=f"nn_measurement_{i-offset}_{i}"))

        sdf_data = th.Variable(dataset.sdf_data_tensor, name="sdf_data")
        sdf_cell_size = th.Variable(dataset.sdf_cell_size, name="sdf_cell_size")
        sdf_origin = th.Variable(dataset.sdf_origin, name="sdf_origin")
        eff_radius = th.Variable(torch.zeros(1, 1), name="eff_radius")

        # -------------------------------------------------------------------- #
        # Creating cost weights
        # -------------------------------------------------------------------- #
        #  - qsp_weight: diagonal cost weight shared across all quasi-static cost
        #    functions.
        #  - mf_between_weight: diagonal cost weight shared across all moving factor
        #    cost functions.
        #  - intersect_weight: scalar cost weight shared across all object-effector
        #    intersection cost functions.
        #  - motion_capture_weight: diagonal cost weight shared across all end-effector
        #    priors cost functions.
        qsp_weight = th.DiagonalCostWeight(
            th.Variable(torch.ones(1, 3), name="qsp_weight")
        )
        mf_between_weight = th.DiagonalCostWeight(
            th.Variable(torch.ones(1, 3), name="mf_between_weight")
        )
        intersect_weight = th.ScaleCostWeight(
            th.Variable(torch.ones(1, 1), name="intersect_weight")
        )
        motion_capture_weight = th.DiagonalCostWeight(
            th.Variable(torch.ones(1, 3), name="mc_weight")
        )

        # -------------------------------------------------------------------- #
        # Creating cost functions
        # -------------------------------------------------------------------- #
        #  - Difference: Penalizes deviation between first object pose from
        #    a global pose prior.
        #  - QuasiStaticPushingPlanar: Penalizes deviation from velocity-only
        #    quasi-static dynamics model QuasiStaticPushingPlanar
        #  - MovingFrameBetween: Penalizes deviation between relative end effector poses
        #    in object frame against a measurement target. Measurement target
        #    `nn_measurements` is obtained from a network prediction.
        #  - EffectorObjectContactPlanar: Penalizes intersections between object and end
        #    effector based on the object sdf.
        #  - Difference: Penalizes deviations of end-effector poses from motion
        #    capture readings

        # Loop over and add all cost functions,
        # cost weights, and their auxiliary variables
        objective = th.Objective()
        nn_meas_idx = 0
        c_square = (np.sqrt(rectangle_shape[0] ** 2 + rectangle_shape[1] ** 2)) ** 2
        for i in range(time_steps):
            if i == 0:
                objective.add(
                    th.Difference(
                        obj_poses[i],
                        obj_start_pose,
                        motion_capture_weight,
                        name=f"obj_priors_{i}",
                    )
                )

            if i < time_steps - 1:
                objective.add(
                    th.eb.QuasiStaticPushingPlanar(
                        obj_poses[i],
                        obj_poses[i + 1],
                        eff_poses[i],
                        eff_poses[i + 1],
                        c_square,
                        qsp_weight,
                        name=f"qsp_{i}",
                    )
                )
            if i >= min_window_moving_frame:
                for offset in range(
                    min_window_moving_frame,
                    np.minimum(i, max_window_moving_frame),
                    step_window_moving_frame,
                ):
                    objective.add(
                        th.eb.MovingFrameBetween(
                            obj_poses[i - offset],
                            obj_poses[i],
                            eff_poses[i - offset],
                            eff_poses[i],
                            nn_measurements[nn_meas_idx],
                            mf_between_weight,
                            name=f"mf_between_{i - offset}_{i}",
                        )
                    )
                    nn_meas_idx = nn_meas_idx + 1

            objective.add(
                th.eb.EffectorObjectContactPlanar(
                    obj_poses[i],
                    eff_poses[i],
                    sdf_origin,
                    sdf_data,
                    sdf_cell_size,
                    eff_radius,
                    intersect_weight,
                    name=f"intersect_{i}",
                )
            )

            objective.add(
                th.Difference(
                    eff_poses[i],
                    motion_captures[i],
                    motion_capture_weight,
                    name=f"eff_priors_{i}",
                )
            )

        if regularization_w > 0.0:
            reg_w = th.ScaleCostWeight(np.sqrt(regularization_w))
            reg_w.to(dtype=torch.double)
            identity_se2 = th.SE2(name="identity")
            for pose_list in [obj_poses, eff_poses]:
                for pose in pose_list:
                    objective.add(
                        th.Difference(
                            pose, identity_se2, reg_w, name=f"reg_{pose.name}"
                        )
                    )

        # -------------------------------------------------------------------- #
        # Creating TheseusLayer
        # -------------------------------------------------------------------- #
        # Wrap the objective and inner-loop optimizer into a `TheseusLayer`.
        # Inner-loop optimizer here is the Levenberg-Marquardt nonlinear optimizer
        # coupled with a dense linear solver based on Cholesky decomposition.
        nl_optimizer = optimizer_cls(
            objective,
            th.CholeskyDenseSolver,
            max_iterations=max_iterations,
            step_size=step_size,
            abs_err_tolerance=0 if force_max_iters else 1e-10,
            rel_err_tolerance=0 if force_max_iters else 1e-8,
        )
        self.theseus_layer = th.TheseusLayer(nl_optimizer)
        self.theseus_layer.to(device=device, dtype=torch.double)

        self.forward = self.theseus_layer.forward

    # Gets a dictionary mapping variable names to tensors, with the batch data needed
    # to update start pose and motion capture data (which is in xytheta format)
    def get_start_pose_and_motion_capture_dict(
        self, batch: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        tensor_dict = {}
        var_names = [self.obj_start_pose.name] + [v.name for v in self.motion_captures]
        for name in var_names:
            tensor_dict[name] = th.SE2(x_y_theta=batch[name]).tensor
        return tensor_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

import theseus.constants

from .lie_group import LieGroup
from .point_types import Point2


class SO2(LieGroup):
    def __init__(
        self,
        theta: Optional[torch.Tensor] = None,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
        strict: bool = False,
    ):
        if theta is not None and tensor is not None:
            raise ValueError("Please provide only one of theta or tensor.")
        if theta is not None:
            dtype = theta.dtype
        super().__init__(tensor=tensor, name=name, dtype=dtype, strict=strict)
        if theta is not None:
            if theta.ndim == 1:
                theta = theta.unsqueeze(1)
            if theta.ndim != 2 or theta.shape[1] != 1:
                raise ValueError(
                    "Argument theta must be have ndim = 1, or ndim=2 and shape[1] = 1."
                )
            self.update_from_angle(theta)

    @staticmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "SO2":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        return SO2.exp_map(
            2
            * theseus.constants.PI
            * torch.rand(
                size[0],
                1,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
            - theseus.constants.PI
        )

    @staticmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "SO2":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        return SO2.exp_map(
            theseus.constants.PI
            * torch.randn(
                size[0],
                1,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    @staticmethod
    def _init_tensor() -> torch.Tensor:  # type: ignore
        return torch.tensor([1.0, 0.0]).view(1, 2)

    def update_from_angle(self, theta: torch.Tensor):
        self.update(torch.cat([theta.cos(), theta.sin()], dim=1))

    def dof(self) -> int:
        return 1

    def __repr__(self) -> str:
        return f"SO2(tensor={self.tensor}, name={self.name})"

    def __str__(self) -> str:
        with torch.no_grad():
            theta = torch.atan2(self[:, 1:], self[:, 0:1])
            return f"SO2(theta={theta}, name={self.name})"

    def theta(self) -> torch.Tensor:
        return self.log_map()

    def _adjoint_impl(self) -> torch.Tensor:
        return torch.ones(self.shape[0], 1, 1, device=self.device, dtype=self.dtype)

    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        self._project_check(euclidean_grad, is_sparse)

        temp = torch.stack((-self[:, 1], self[:, 0]), dim=1)

        if is_sparse:
            return torch.einsum("i...k,i...k->i...", euclidean_grad, temp).unsqueeze(-1)
        else:
            return torch.einsum("...k,...k", euclidean_grad, temp).unsqueeze(-1)

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        with torch.no_grad():
            if tensor.ndim != 2 or tensor.shape[1] != 2:
                raise ValueError("SO2 data tensors can only be 2D vectors.")

            MATRIX_EPS = theseus.constants._SO2_MATRIX_EPS[tensor.dtype]
            if tensor.dtype != torch.float64:
                tensor = tensor.double()

            _check = (
                torch.linalg.norm(tensor, dim=1) - 1
            ).abs().max().item() <= MATRIX_EPS

        return _check

    @staticmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "SO2":
        so2 = SO2(dtype=tangent_vector.dtype)
        so2.update_from_angle(tangent_vector)

        if jacobians is not None:
            SO2._check_jacobians_list(jacobians)
            jacobians.append(
                torch.ones(
                    tangent_vector.shape[0],
                    1,
                    1,
                    dtype=tangent_vector.dtype,
                    device=tangent_vector.device,
                )
            )

        return so2

    @staticmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        if tensor.ndim != 2 or tensor.shape[1] != 2:
            raise ValueError("SO2 data tensors can only be 2D vectors.")

        data_norm = torch.norm(tensor, dim=1, keepdim=True)
        near_zero = data_norm < theseus.constants._SO2_NORMALIZATION_EPS[tensor.dtype]
        data_norm_nz = torch.where(
            near_zero,
            torch.tensor(1.0, dtype=tensor.dtype, device=tensor.device),
            data_norm,
        )
        default_data = torch.tensor(
            [1, 0], dtype=tensor.dtype, device=tensor.device
        ).expand([tensor.shape[0], 2])
        return torch.where(near_zero, default_data, tensor / data_norm_nz)

    def _log_map_impl(
        self, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        if jacobians is not None:
            SO2._check_jacobians_list(jacobians)
            jacobians.append(
                torch.ones(
                    self.shape[0],
                    1,
                    1,
                    dtype=self.dtype,
                    device=self.device,
                )
            )

        cosine, sine = self.to_cos_sin()
        return torch.atan2(sine, cosine).unsqueeze(1)

    def _compose_impl(self, so2_2: LieGroup) -> "SO2":
        so2_2 = cast(SO2, so2_2)
        cos_1, sin_1 = self.to_cos_sin()
        cos_2, sin_2 = so2_2.to_cos_sin()
        new_cos = cos_1 * cos_2 - sin_1 * sin_2
        new_sin = sin_1 * cos_2 + cos_1 * sin_2
        return SO2(tensor=torch.stack([new_cos, new_sin], dim=1), strict=False)

    def _inverse_impl(self, get_jacobian: bool = False) -> "SO2":
        cosine, sine = self.to_cos_sin()
        return SO2(tensor=torch.stack([cosine, -sine], dim=1), strict=False)

    def _rotate_shape_check(self, point: Union[Point2, torch.Tensor]):
        err_msg = (
            f"SO2 can only transform vectors of shape [{self.shape[0]}, 2] or [1, 2], "
            f"but the input has shape {point.shape}."
        )

        if isinstance(point, torch.Tensor):
            if not point.ndim == 2 or point.shape[1] != 2:
                raise ValueError(err_msg)
        elif point.dof() != 2:
            raise ValueError(err_msg)
        if (
            point.shape[0] != self.shape[0]
            and point.shape[0] != 1
            and self.shape[0] != 1
        ):
            raise ValueError(
                "Input point batch size is not broadcastable with group batch size."
            )

    @staticmethod
    def _rotate_from_cos_sin(
        point: Union[Point2, torch.Tensor],
        cosine: torch.Tensor,
        sine: torch.Tensor,
    ) -> Point2:
        batch_size = max(point.shape[0], cosine.shape[0])
        if isinstance(point, torch.Tensor):
            if point.ndim != 2 or point.shape[1] != 2:
                raise ValueError(
                    f"Point tensor must have shape batch_size x 2, "
                    f"but received {point.shape}."
                )
            point_tensor = point
        else:
            point_tensor = point.tensor
        px, py = point_tensor[:, 0], point_tensor[:, 1]
        new_point_tensor = torch.empty(
            batch_size, 2, device=cosine.device, dtype=cosine.dtype
        )
        new_point_tensor[:, 0] = cosine * px - sine * py
        new_point_tensor[:, 1] = sine * px + cosine * py
        return Point2(tensor=new_point_tensor)

    def rotate(
        self,
        point: Union[Point2, torch.Tensor],
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> Point2:
        self._rotate_shape_check(point)
        cosine, sine = self.to_cos_sin()
        ret = SO2._rotate_from_cos_sin(point, cosine, sine)
        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            Jrot = torch.stack([-ret.y(), ret.x()], dim=1).view(-1, 2, 1)
            Jpnt = self.to_matrix().expand(ret.shape[0], -1, -1)
            jacobians.extend([Jrot, Jpnt])
        return ret

    def unrotate(
        self,
        point: Union[Point2, torch.Tensor],
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> Point2:
        self._rotate_shape_check(point)
        cosine, sine = self.to_cos_sin()
        ret = SO2._rotate_from_cos_sin(point, cosine, -sine)
        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            Jrot = torch.stack([ret.y(), -ret.x()], dim=1).view(-1, 2, 1)
            Jpnt = self.to_matrix().transpose(2, 1).expand(ret.shape[0], -1, -1)
            jacobians.extend([Jrot, Jpnt])
        return ret

    def to_cos_sin(self) -> Tuple[torch.Tensor, torch.Tensor]:
        return self[:, 0], self[:, 1]

    def to_matrix(self) -> torch.Tensor:
        matrix = torch.empty(self.shape[0], 2, 2).to(
            device=self.device, dtype=self.dtype
        )
        cosine, sine = self.to_cos_sin()
        matrix[:, 0, 0] = cosine
        matrix[:, 0, 1] = -sine
        matrix[:, 1, 0] = sine
        matrix[:, 1, 1] = cosine
        return matrix

    @staticmethod
    def hat(tangent_vector: torch.Tensor) -> torch.Tensor:
        matrix = torch.zeros(tangent_vector.shape[0], 2, 2).to(
            dtype=tangent_vector.dtype,
            device=tangent_vector.device,
        )
        matrix[:, 0, 1] = -tangent_vector.view(-1)
        matrix[:, 1, 0] = tangent_vector.view(-1)
        return matrix

    @staticmethod
    def vee(matrix: torch.Tensor) -> torch.Tensor:
        _check = matrix.ndim == 3 and matrix.shape[1:] == (2, 2)
        _check &= matrix[:, 0, 0].abs().max().item() < theseus.constants.EPS
        _check &= matrix[:, 1, 1].abs().max().item() < theseus.constants.EPS
        _check &= torch.allclose(matrix[:, 0, 1], -matrix[:, 1, 0])
        if not _check:
            raise ValueError("Invalid hat matrix for SO2.")
        return matrix[:, 1, 0].clone().view(-1, 1)

    def _copy_impl(self, new_name: Optional[str] = None) -> "SO2":
        return SO2(tensor=self.tensor.clone(), name=new_name)

    # only added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "SO2":
        return cast(SO2, super().copy(new_name=new_name))


rand_so2 = SO2.rand
randn_so2 = SO2.randn



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc

import numpy as np
from scipy.sparse import csc_matrix, csr_matrix


class SparseStructure(abc.ABC):
    def __init__(
        self,
        col_ind: np.ndarray,
        row_ptr: np.ndarray,
        num_rows: int,
        num_cols: int,
        dtype: np.dtype = np.float_,
    ):
        self.col_ind = col_ind
        self.row_ptr = row_ptr
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.dtype = dtype

    def csr_straight(self, val):
        return csr_matrix(
            (val, self.col_ind, self.row_ptr),
            (self.num_rows, self.num_cols),
            dtype=self.dtype,
        )

    def csc_transpose(self, val):
        return csc_matrix(
            (val, self.col_ind, self.row_ptr),
            (self.num_cols, self.num_rows),
            dtype=self.dtype,
        )

    def mock_csc_transpose(self):
        return csc_matrix(
            (np.ones(len(self.col_ind), dtype=self.dtype), self.col_ind, self.row_ptr),
            (self.num_cols, self.num_rows),
            dtype=self.dtype,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, List, Optional, Type

import torch

from theseus.core import Objective
from theseus.optimizer import Linearization, SparseLinearization
from theseus.optimizer.autograd import LUCudaSolveFunction

from .linear_solver import LinearSolver


class LUCudaSparseSolver(LinearSolver):
    def __init__(
        self,
        objective: Objective,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        num_solver_contexts=1,
        batch_size: Optional[int] = None,
        auto_reset: bool = True,
        **kwargs,
    ):
        if not torch.cuda.is_available():
            raise RuntimeError("Cuda not available, LUCudaSparseSolver cannot be used")

        linearization_cls = linearization_cls or SparseLinearization
        if not linearization_cls == SparseLinearization:
            raise RuntimeError(
                "LUCudaSparseSolver only works with theseus.optimizer.SparseLinearization,"
                + f" got {type(self.linearization)}"
            )

        super().__init__(objective, linearization_cls, linearization_kwargs, **kwargs)
        self.linearization: SparseLinearization = self.linearization

        self._num_solver_contexts: int = num_solver_contexts

        if self.linearization.structure().num_rows:
            if batch_size is not None:
                self.reset(batch_size=batch_size)
            else:
                self.reset()

        self._objective = objective
        self._auto_reset = auto_reset

    def reset(self, batch_size: int = 16):
        if not torch.cuda.is_available():
            raise RuntimeError("Cuda not available, LUCudaSparseSolver cannot be used")

        try:
            from theseus.extlib.cusolver_lu_solver import CusolverLUSolver
        except Exception as e:
            raise RuntimeError(
                "Theseus C++/Cuda extension cannot be loaded\n"
                "even if Cuda appears to be available. Make sure Theseus\n"
                "is installed with Cuda support (export CUDA_HOME=...)\n"
                f"{type(e).__name__}: {e}"
            )

        self.A_rowPtr = torch.tensor(
            self.linearization.structure().row_ptr, dtype=torch.int32
        ).cuda()
        self.A_colInd = torch.tensor(
            self.linearization.structure().col_ind, dtype=torch.int32
        ).cuda()
        At_mock = self.linearization.structure().mock_csc_transpose()
        AtA_mock = (At_mock @ At_mock.T).tocsr()

        # symbolic decomposition depending on the sparse structure, done with mock data
        # HACK: we generate several context, as by cublas the symbolic_decomposition is
        # also a context for factorization, and the two cannot be separated
        AtA_rowPtr = torch.tensor(AtA_mock.indptr, dtype=torch.int32).cuda()
        AtA_colInd = torch.tensor(AtA_mock.indices, dtype=torch.int32).cuda()
        self._solver_contexts: List[CusolverLUSolver] = [
            CusolverLUSolver(
                batch_size,
                AtA_mock.shape[1],
                AtA_rowPtr,
                AtA_colInd,
            )
            for _ in range(self._num_solver_contexts)
        ]
        self._last_solver_context: int = self._num_solver_contexts - 1

    def solve(
        self,
        damping: Optional[float] = None,
        ellipsoidal_damping: bool = True,
        damping_eps: float = 1e-8,
        **kwargs,
    ) -> torch.Tensor:
        if self._auto_reset:
            if self._solver_contexts[0].batch_size != self._objective.batch_size:
                self.reset(self._objective.batch_size)
        if not isinstance(self.linearization, SparseLinearization):
            raise RuntimeError(
                "CholmodSparseSolver only works with theseus.optimizer.SparseLinearization."
            )

        self._last_solver_context = (
            self._last_solver_context + 1
        ) % self._num_solver_contexts

        if damping is None:
            damping_alpha_beta = None
        else:
            # See Nocedal and Wright, Numerical Optimization, pp. 260 and 261
            # https://www.csie.ntu.edu.tw/~r97002/temp/num_optimization.pdf
            damping_alpha_beta = (
                (damping, damping_eps) if ellipsoidal_damping else (0.0, damping)
            )

        return LUCudaSolveFunction.apply(
            self.linearization.A_val,
            self.linearization.b,
            self.linearization.structure(),
            self.A_rowPtr,
            self.A_colInd,
            self._solver_contexts[self._last_solver_context],
            damping_alpha_beta,
            True,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional

import numpy as np
import torch

from theseus.core import Objective

from .linear_system import SparseStructure
from .linearization import Linearization
from .variable_ordering import VariableOrdering


class SparseLinearization(Linearization):
    def __init__(
        self,
        objective: Objective,
        ordering: Optional[VariableOrdering] = None,
        **kwargs
    ):
        super().__init__(objective, ordering)

        # we prepare the indices for At as csc matrix (or A as csr, same thing)
        # for similarity with dense_linearization code we build A as csr, then
        # actually we have At as csc and can feed it to `cholesky_AAt` routine
        # we want a unique set of rowPtr/colInd indices for A for all batches
        # we also save pointers to the data block, so that we can later quickly
        # write the data blocks
        A_col_ind: List[int] = []
        A_row_ptr: List[int] = [0]

        # ptr to data block (stride = sum of variable.dim()
        cost_function_block_pointers = []
        cost_function_row_block_starts = []  # where data start for this row block
        cost_function_stride = []  # total jacobian cols

        for _, cost_function in enumerate(self.objective._get_iterator()):
            num_rows = cost_function.dim()
            col_slices_indices = []
            for var_idx_in_cost_function, variable in enumerate(
                cost_function.optim_vars
            ):
                var_idx_in_order = self.ordering.index_of(
                    cost_function.optim_var_at(var_idx_in_cost_function).name
                )
                var_start_col = self.var_start_cols[var_idx_in_order]
                num_cols = variable.dof()
                col_slice = slice(var_start_col, var_start_col + num_cols)
                col_slices_indices.append((col_slice, var_idx_in_cost_function))

            # sort according to how they will be written inside A
            col_slices_indices.sort()
            sorted_block_sizes = [(s.stop - s.start) for s, _ in col_slices_indices]
            sorted_block_pointers = np.cumsum([0] + sorted_block_sizes)[:-1]
            sorted_indices = np.array([i for _, i in col_slices_indices])
            block_pointers: np.ndarray = np.ndarray(
                (len(col_slices_indices),), dtype=int
            )
            block_pointers[sorted_indices] = sorted_block_pointers
            cost_function_block_pointers.append(block_pointers)

            cost_function_row_block_starts.append(len(A_col_ind))
            col_ind = [c for s, _ in col_slices_indices for c in range(s.start, s.stop)]
            cost_function_stride.append(len(col_ind))

            for _ in range(num_rows):
                A_col_ind += col_ind
                A_row_ptr.append(len(A_col_ind))

        # not batched, these data are the same across batches
        self.cost_function_block_pointers = cost_function_block_pointers
        self.cost_function_row_block_starts: np.ndarray = np.array(
            cost_function_row_block_starts, dtype=int
        )
        self.cost_function_stride: np.ndarray = np.array(
            cost_function_stride, dtype=int
        )
        self.A_row_ptr: np.ndarray = np.array(A_row_ptr, dtype=int)
        self.A_col_ind: np.ndarray = np.array(A_col_ind, dtype=int)

        # batched data
        self.A_val: torch.Tensor = None
        self.b: torch.Tensor = None
        self.Atb: torch.Tensor = None

    def _linearize_jacobian_impl(self):
        # those will be fully overwritten, no need to zero:
        self.A_val = torch.empty(
            size=(self.objective.batch_size, len(self.A_col_ind)),
            device=self.objective.device,
            dtype=self.objective.dtype,
        )
        self.b = torch.empty(
            size=(self.objective.batch_size, self.num_rows),
            device=self.objective.device,
            dtype=self.objective.dtype,
        )

        err_row_idx = 0
        for f_idx, cost_function in enumerate(self.objective._get_iterator()):
            jacobians, error = cost_function.weighted_jacobians_error()
            num_rows = cost_function.dim()
            row_slice = slice(err_row_idx, err_row_idx + num_rows)

            # we will view the blocks of rows inside `A_val` as `num_rows` x `stride` matrix
            block_start = self.cost_function_row_block_starts[f_idx]
            stride = self.cost_function_stride[f_idx]
            block = self.A_val[:, block_start : block_start + stride * num_rows].view(
                -1, num_rows, stride
            )
            block_pointers = self.cost_function_block_pointers[f_idx]

            for var_idx_in_cost_function, var_jacobian in enumerate(jacobians):

                # the proper block is written, using the precomputed index in `block_pointers`
                num_cols = var_jacobian.shape[2]
                pointer = block_pointers[var_idx_in_cost_function]
                block[:, :, pointer : pointer + num_cols] = var_jacobian

            self.b[:, row_slice] = -error
            err_row_idx += cost_function.dim()

    def structure(self):
        return SparseStructure(
            self.A_col_ind,
            self.A_row_ptr,
            self.num_rows,
            self.num_cols,
            dtype=np.float64 if self.objective.dtype == torch.double else np.float32,
        )

    def _linearize_hessian_impl(self):
        self._linearize_jacobian_impl()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .misc import TrajectoryDataset, generate_trajectory_figs
from .models import (
    InitialTrajectoryModel,
    ScalarCollisionWeightAndCostEpstModel,
    ScalarCollisionWeightModel,
)
from .motion_planner import MotionPlanner



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
from torch.autograd.function import once_differentiable

from theseus.core import (
    CostFunction,
    CostWeight,
    Objective,
    ScaleCostWeight,
    Variable,
    Vectorize,
)
from theseus.geometry import LieGroup, Manifold
from theseus.optimizer import Optimizer, OptimizerInfo
from theseus.optimizer.linear import LinearSolver
from theseus.optimizer.nonlinear import BackwardMode, GaussNewton


class TheseusLayer(nn.Module):
    def __init__(
        self,
        optimizer: Optimizer,
        vectorize: bool = True,
        empty_cuda_cache: bool = False,
    ):
        super().__init__()
        self.objective = optimizer.objective
        if vectorize and not self.objective.vectorized:
            Vectorize(self.objective, empty_cuda_cache=empty_cuda_cache)
        self.optimizer = optimizer
        self._objectives_version = optimizer.objective.current_version
        self._dlm_bwd_objective = None
        self._dlm_bwd_optimizer = None

    def forward(
        self,
        input_tensors: Optional[Dict[str, torch.Tensor]] = None,
        optimizer_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Dict[str, torch.Tensor], OptimizerInfo]:
        if self._objectives_version != self.objective.current_version:
            raise RuntimeError(
                "The objective was modified after the layer's construction, which is "
                "currently not supported."
            )
        optimizer_kwargs = optimizer_kwargs or {}
        backward_mode = optimizer_kwargs.get("backward_mode", None)
        if backward_mode == BackwardMode.DLM:
            dlm_epsilon = optimizer_kwargs.get(
                TheseusLayerDLMForward._DLM_EPSILON_STR, 1e-2
            )
            if not isinstance(dlm_epsilon, float):
                raise ValueError(
                    f"{TheseusLayerDLMForward._DLM_EPSILON_STR} must be a float "
                    f"but {type(dlm_epsilon)} was given."
                )

            if self._dlm_bwd_objective is None:
                _obj, _opt = _instantiate_dlm_bwd_objective(self.objective)
                _obj.to(self.device)
                self._dlm_bwd_objective = _obj
                self._dlm_bwd_optimizer = _opt

            # Tensors cannot be passed inside containers, else we run into memory leaks.
            input_keys, input_vals = zip(*input_tensors.items())
            differentiable_tensors = [t for t in input_vals if t.requires_grad]

            *vars, info = TheseusLayerDLMForward.apply(
                self.objective,
                self.optimizer,
                optimizer_kwargs,
                self._dlm_bwd_objective,
                self._dlm_bwd_optimizer,
                dlm_epsilon,
                len(input_keys),
                *input_keys,
                *input_vals,
                *differentiable_tensors,
            )
        else:
            vars, info = _forward(
                self.objective, self.optimizer, optimizer_kwargs, input_tensors
            )
        values = dict(zip(self.objective.optim_vars.keys(), vars))
        return values, info

    def compute_samples(
        self,
        linear_solver: LinearSolver = None,
        n_samples: int = 10,
        temperature: float = 1.0,
    ) -> torch.Tensor:
        # When samples are not available, return None. This makes the outer learning loop default
        # to a perceptron loss using the mean trajectory solution from the optimizer.
        if linear_solver is None:
            return None

        # Sampling from multivariate normal using a Cholesky decomposition of AtA,
        # http://www.statsathome.com/2018/10/19/sampling-from-multivariate-normal-precision-and-covariance-parameterizations/
        delta = linear_solver.solve()
        AtA = linear_solver.linearization.hessian_approx() / temperature
        sqrt_AtA = torch.linalg.cholesky(AtA).permute(0, 2, 1)

        batch_size, n_vars = delta.shape
        y = torch.normal(
            mean=torch.zeros((n_vars, n_samples), device=delta.device),
            std=torch.ones((n_vars, n_samples), device=delta.device),
        )
        delta_samples = (torch.triangular_solve(y, sqrt_AtA).solution) + (
            delta.unsqueeze(-1)
        ).repeat(1, 1, n_samples)

        x_samples = torch.zeros((batch_size, n_vars, n_samples), device=delta.device)
        for sidx in range(0, n_samples):
            var_idx = 0
            for var in linear_solver.linearization.ordering:
                new_var = var.retract(
                    delta_samples[:, var_idx : var_idx + var.dof(), sidx]
                )
                x_samples[:, var_idx : var_idx + var.dof(), sidx] = new_var.tensor
                var_idx = var_idx + var.dof()

        return x_samples

    # Applies to() with given args to all tensors in the objective
    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)
        self.objective.to(*args, **kwargs)

    @property
    def device(self) -> torch.device:
        return self.objective.device

    @property
    def dtype(self) -> torch.dtype:
        return self.objective.dtype


def _forward(objective, optimizer, optimizer_kwargs, input_tensors):
    objective.update(input_tensors)
    info = optimizer.optimize(**optimizer_kwargs)
    vars = [var.tensor for var in objective.optim_vars.values()]
    return vars, info


class TheseusLayerDLMForward(torch.autograd.Function):
    """
    Functionally the same as the forward method in a TheseusLayer
    but computes the direct loss minimization in the backward pass.
    """

    _DLM_EPSILON_STR = "dlm_epsilon"
    _GRAD_SUFFIX = "_grad"

    @staticmethod
    def forward(
        ctx,
        objective,
        optimizer,
        optimizer_kwargs,
        bwd_objective,
        bwd_optimizer,
        epsilon,
        n,
        *inputs,
    ):
        input_keys = inputs[:n]
        input_vals = inputs[n : 2 * n]
        differentiable_tensors = inputs[2 * n :]
        ctx.n = n
        ctx.k = len(differentiable_tensors)

        inputs = dict(zip(input_keys, input_vals))
        ctx.input_keys = input_keys

        optim_tensors, info = _forward(objective, optimizer, optimizer_kwargs, inputs)

        # Skip computation if there are no differentiable inputs.
        if ctx.k > 0:
            ctx.bwd_objective = bwd_objective
            ctx.bwd_optimizer = bwd_optimizer
            ctx.epsilon = epsilon

            # Precompute and cache this.
            with torch.enable_grad():
                grad_sol = torch.autograd.grad(
                    objective.error_squared_norm().sum(),
                    differentiable_tensors,
                    allow_unused=True,
                )
            ctx.save_for_backward(
                *input_vals, *grad_sol, *differentiable_tensors, *optim_tensors
            )
        return (*optim_tensors, info)

    @staticmethod
    @once_differentiable
    def backward(ctx, *grad_outputs):
        n, k = ctx.n, ctx.k
        saved_tensors = ctx.saved_tensors
        input_vals = saved_tensors[:n]
        grad_sol = saved_tensors[n : n + k]
        differentiable_tensors = saved_tensors[n + k : n + k + k]
        optim_tensors = saved_tensors[n + k + k :]
        grad_outputs = grad_outputs[:-1]

        bwd_objective = ctx.bwd_objective
        bwd_optimizer = ctx.bwd_optimizer
        epsilon = ctx.epsilon
        input_keys = ctx.input_keys

        # Update the optim vars to their solutions.
        bwd_data = dict(zip(input_keys, input_vals))
        for k, v in zip(bwd_objective.optim_vars.keys(), optim_tensors):
            bwd_data[k] = v.detach()

        # Add in gradient values.
        grad_data = {
            TheseusLayerDLMForward._DLM_EPSILON_STR: torch.tensor(epsilon)
            .to(grad_outputs[0])
            .reshape(1, 1)
        }
        for i, name in enumerate(bwd_objective.optim_vars.keys()):
            grad_data[name + TheseusLayerDLMForward._GRAD_SUFFIX] = grad_outputs[i]
        bwd_data.update(grad_data)

        # Solve backward objective.
        bwd_objective.update(bwd_data)
        with torch.no_grad():
            bwd_optimizer.linear_solver.linearization.linearize()
            delta = bwd_optimizer.linear_solver.solve()
            bwd_optimizer.objective.retract_optim_vars(
                delta, bwd_optimizer.linear_solver.linearization.ordering
            )

        # Compute gradients.
        with torch.enable_grad():
            grad_perturbed = torch.autograd.grad(
                bwd_objective.error_squared_norm().sum(),
                differentiable_tensors,
                allow_unused=True,
            )

        nones = [None] * (ctx.n * 2)
        grads = [
            (gs - gp) / epsilon if gs is not None else None
            for gs, gp in zip(grad_sol, grad_perturbed)
        ]
        return (None, None, None, None, None, None, None, *nones, *grads)


class _DLMPerturbation(CostFunction):
    def __init__(
        self,
        var: Manifold,
        epsilon: Variable,
        grad: Variable,
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        if not isinstance(var, LieGroup):
            raise ValueError(
                f"DLM requires LieGroup-type variables, but "
                f"{var.name} has type {var.__class__.__name__}"
            )
        super().__init__(cost_weight, name=name)
        assert epsilon.ndim == 2 and epsilon.shape[1] == 1
        self.var = var
        self.epsilon = epsilon
        self.grad = grad
        self.register_optim_var("var")
        self.register_aux_vars(["epsilon", "grad"])

    def error(self) -> torch.Tensor:
        err = (
            self.epsilon.tensor.view((-1,) + (1,) * (self.var.ndim - 1))
            * self.var.tensor
            - 0.5 * self.grad.tensor
        )
        return err.flatten(start_dim=1)

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        d = self.dim()
        aux = (
            torch.eye(d, dtype=self.epsilon.dtype, device=self.epsilon.device)
            .unsqueeze(0)
            .expand(self.var.shape[0], d, d)
        )
        euclidean_grad_flat = self.epsilon.tensor.view(-1, 1, 1) * aux
        euclidean_grad = euclidean_grad_flat.unflatten(2, self.var.shape[1:])
        return [self.var.project(euclidean_grad, is_sparse=True)], self.error()

    def dim(self) -> int:
        return np.prod(self.var.tensor.shape[1:])

    def _copy_impl(self, new_name: Optional[str] = None) -> "CostFunction":
        return _DLMPerturbation(
            self.var.copy(),
            self.epsilon.copy(),
            self.grad.copy(),
            self.weight.copy(),
            name=new_name,
        )


def _instantiate_dlm_bwd_objective(objective: Objective):
    bwd_objective = objective.copy()
    epsilon_var = Variable(
        torch.ones(1, 1, dtype=bwd_objective.dtype, device=bwd_objective.device),
        name=TheseusLayerDLMForward._DLM_EPSILON_STR,
    )
    unit_weight = ScaleCostWeight(1.0)
    unit_weight.to(dtype=objective.dtype, device=objective.device)
    for name, var in bwd_objective.optim_vars.items():
        grad_var = Variable(
            torch.zeros_like(var.tensor),
            name=name + TheseusLayerDLMForward._GRAD_SUFFIX,
        )
        bwd_objective.add(
            _DLMPerturbation(
                var, epsilon_var, grad_var, unit_weight, name="dlm_perturbation" + name
            )
        )

    bwd_optimizer = GaussNewton(
        bwd_objective,
        max_iterations=1,
        step_size=1.0,
    )
    return bwd_objective, bwd_optimizer



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Tuple

import torch


# ------------------------------------------------------------ #
# --------------------------- LOSSES ------------------------- #
# ------------------------------------------------------------ #
def soft_loss_cauchy(
    x: torch.Tensor, radius: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    ratio = (x + radius) / radius
    val = torch.log(ratio) * radius
    der = 1.0 / ratio
    return val, der


def soft_loss_huber_like(
    x: torch.Tensor, radius: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    ratio = (x + radius) / radius
    sq = torch.sqrt(ratio)
    val = (sq - 1) * radius
    der = 0.5 / sq
    return val, der


# ------------------------------------------------------------ #
# ----------------------------- RNG -------------------------- #
# ------------------------------------------------------------ #
# returns a uniformly random point of the 2-sphere
def random_s2(dtype: torch.dtype = torch.float64) -> torch.Tensor:
    theta = torch.rand(()) * math.tau
    z = torch.rand(()) * 2 - 1
    r = torch.sqrt(1 - z**2)
    return torch.tensor([r * torch.cos(theta), r * torch.sin(theta), z]).to(dtype=dtype)


# returns a uniformly random point of the 3-sphere
def random_s3(dtype: torch.dtype = torch.float64) -> torch.Tensor:
    u, v, w = torch.rand(3)
    return torch.tensor(
        [
            torch.sqrt(1 - u) * torch.sin(math.tau * v),
            torch.sqrt(1 - u) * torch.cos(math.tau * v),
            torch.sqrt(u) * torch.sin(math.tau * w),
            torch.sqrt(u) * torch.cos(math.tau * w),
        ]
    ).to(dtype=dtype)


def random_small_quaternion(
    max_degrees: float, min_degrees: int = 0, dtype: torch.dtype = torch.float64
) -> torch.Tensor:
    x, y, z = random_s2(dtype=dtype)
    theta = (
        (min_degrees + (max_degrees - min_degrees) * torch.rand((), dtype=dtype))
        * math.tau
        / 360.0
    )
    c, s = torch.cos(theta), torch.sin(theta)
    return torch.tensor([c, s * x, s * y, s * z])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import threading
from typing import Any

import torch

from .lie_group import LieGroup


class _LieGroupContext(object):
    contexts = threading.local()

    @classmethod
    def get_context(cls):
        if not hasattr(cls.contexts, "use_lie_tangent"):
            cls.contexts.use_lie_tangent = False
        return cls.contexts.use_lie_tangent

    @classmethod
    def set_context(cls, use_lie_tangent: bool):
        cls.contexts.use_lie_tangent = use_lie_tangent


class set_lie_tangent_enabled(object):
    def __init__(self, mode: bool) -> None:
        self.prev = _LieGroupContext.get_context()
        _LieGroupContext.set_context(mode)

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        _LieGroupContext.set_context(self.prev)


class enable_lie_tangent(object):
    def __enter__(self) -> None:
        self.prev = _LieGroupContext.get_context()
        _LieGroupContext.set_context(True)

    def __exit__(self, typ, value, traceback) -> None:
        _LieGroupContext.set_context(self.prev)


class no_lie_tangent(_LieGroupContext):
    def __enter__(self):
        self.prev = super().get_context()
        _LieGroupContext.set_context(False)
        return self

    def __exit__(self, typ, value, traceback):
        _LieGroupContext.set_context(self.prev)


class LieGroupTensor(torch.Tensor):
    from torch._C import _disabled_torch_function_impl  # type: ignore

    __torch_function__ = _disabled_torch_function_impl

    def __new__(cls, group):
        return torch.Tensor._make_subclass(cls, group.tensor)

    def __init__(self, group: LieGroup):
        self.group_cls = type(group)

    def add_(self, update, alpha=1):
        if _LieGroupContext.get_context():
            group = self.group_cls(tensor=self.data)
            grad = group.project(update)
            self.set_(group.retract(alpha * grad).tensor)
        else:
            self.add_(update, alpha=alpha)

        return self

    def addcdiv_(self, tensor1, tensor2, value=1):
        self.add_(
            value * tensor1 / tensor2
        ) if _LieGroupContext.get_context() else super().addcdiv_(
            tensor1, tensor2, value=value
        )
        return self

    def addcmul_(self, tensor1, tensor2, value=1):
        self.add_(
            value * tensor1 * tensor2
        ) if _LieGroupContext.get_context() else super().addcmul_(
            tensor1, tensor2, value=value
        )
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from ..linear_system import SparseStructure


class LUCudaSolveFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, *args, **kwargs):
        if not torch.cuda.is_available():
            raise RuntimeError("Cuda not available, LUCudaSolveFunction cannot be used")

        try:
            from theseus.extlib.cusolver_lu_solver import CusolverLUSolver
            from theseus.extlib.mat_mult import apply_damping, mult_MtM, tmat_vec
        except Exception as e:
            raise RuntimeError(
                "Theseus C++/Cuda extension cannot be loaded\n"
                "even if Cuda appears to be available. Make sure Theseus\n"
                "is installed with Cuda support (export CUDA_HOME=...)\n"
                f"{type(e).__name__}: {e}"
            )

        A_val: torch.Tensor = args[0]
        b: torch.Tensor = args[1]
        sparse_structure: SparseStructure = args[2]
        A_rowPtr: torch.Tensor = args[3]
        A_colInd: torch.Tensor = args[4]
        solver_context: CusolverLUSolver = args[5]
        damping_alpha_beta: float = args[6]
        check_factor_id: bool = args[7]

        AtA_rowPtr = solver_context.A_rowPtr
        AtA_colInd = solver_context.A_colInd

        batch_size = A_val.shape[0]

        AtA = mult_MtM(batch_size, A_rowPtr, A_colInd, A_val, AtA_rowPtr, AtA_colInd)
        if damping_alpha_beta is not None:
            AtA_args = sparse_structure.num_cols, AtA_rowPtr, AtA_colInd, AtA
            apply_damping(batch_size, *AtA_args, *damping_alpha_beta)
        solver_context.factor(AtA)

        A_args = sparse_structure.num_cols, A_rowPtr, A_colInd, A_val
        Atb = tmat_vec(batch_size, *A_args, b)
        x = Atb.clone()
        solver_context.solve(x)  # solve in place

        ctx.b = b
        ctx.x = x
        ctx.A_val = A_val
        ctx.A_rowPtr = A_rowPtr
        ctx.A_colInd = A_colInd
        ctx.sparse_structure = sparse_structure
        ctx.solver_context = solver_context
        ctx.damping_alpha_beta = damping_alpha_beta

        # HACK: allows to check if the context has been reused (and overwritten)
        ctx.factor_id = solver_context.factor_id if check_factor_id else None

        return x

    # Let v row vector, and w column vector of dimension n, m, and
    # A an nxm matrix. Then
    #   v * A * w = Sum(v_i * A_ij * w_j) = [v (X) w] . A
    # Where by [v (X) w] we mean the nxm matrix which is the
    # tensor product of v and w, and "." is the componentwise
    # dot product of the two nxm matrices.
    #
    # Now, we have
    #      At * A * x = At * b
    #
    # Therefore if A, b, x are parametrized (eg. by u) we have deriving
    #
    # (i)  At'*A*x + At*A'*x + At*A*x' = At'*b + At*b'
    #
    # indicating A'=dA/du, b'=db/du, x'=dx/du
    #
    # Now, assume we have a function f of x, and G = df/dx be the
    # gradient that we consider a row vector, so G*x' is df/du.
    #
    # To compute df/db and df/dA, make x' explicit in the (i):
    #
    #      x' = (At * A)^{-1} (At*b' + At'*b - At'*A*x - At*A'*x)
    #
    # So multiplying by the row vector G we have
    #
    #   G*x' = G*(At*A)^{-1}*At * b' + G*(At*A)^{-1} * (At'*b - At'*A*x - At*A'*x)
    #        = H * At * b' + H * At' * (b - A*x) - H * At * A' * x
    #        = (H*At) * b' + [H (X) (b-A*x)] . At' - [H*At (X) x] . A'
    #        = (H*At) * b' + [(b-A*x) (X) H - H*At (X) x] . A'
    # after putting H = G*(At*A)^{-1} for convenience.
    #
    # Therefore after switching to column vectors we have
    #   df/db = A*H
    # (where H = (At*A)^{-1}*G), while
    #   df/dA = (b-A*x) (X) H - A*H (X) x
    # The two tensor products means that to compute the gradient of
    # a block of A we have to multiply entries taken from (b-A*x) and H,
    # and blocks taken from A*H and x.
    #
    # Here we assume we are provided x and H after the linear solver
    # has been applied to Atb and the gradient G.

    # With (large) multiplicative damping, as above with extra terms:
    #      x' = ... - (AtA_damped)^{-1} * alpha*AtA_diag'*x
    # So multiplying by the row vector G we have
    #      G*x' = ... - H * alpha * AtA_diag' * x
    # Note that '...' the part multiplying H[j]*alpha*x[j] is AtA_diag'[j], ie
    # 2 times the scalar product of A's an (A')'s j-th colum. Therefore
    # (A')'s j-th colum is multiplying A's j-th colum by 2*H[j]*alpha*x[j]
    @staticmethod
    def backward(ctx, grad_output):

        if not torch.cuda.is_available():
            raise RuntimeError("Cuda not available, LUCudaSolveFunction cannot be used")

        try:
            from theseus.extlib.mat_mult import mat_vec
        except Exception as e:
            raise RuntimeError(
                "Theseus C++/Cuda extension cannot be loaded\n"
                "even if Cuda appears to be available. Make sure Theseus\n"
                "is installed with Cuda support (export CUDA_HOME=...)\n"
                f"{type(e).__name__}: {e}"
            )

        # HACK: check if the context has been reused (and overwritten)
        if ctx.factor_id is not None and ctx.factor_id != ctx.solver_context.factor_id:
            raise RuntimeError(
                "Factoring context was overwritten! Increase the number of contexts"
            )

        batch_size = grad_output.shape[0]
        targs = {"dtype": grad_output.dtype, "device": "cuda"}  # grad_output.device}

        H = grad_output.clone()
        ctx.solver_context.solve(H)  # solve in place

        A_args = ctx.sparse_structure.num_cols, ctx.A_rowPtr, ctx.A_colInd, ctx.A_val
        AH = mat_vec(batch_size, *A_args, H)
        b_Ax = ctx.b - mat_vec(batch_size, *A_args, ctx.x)

        # now we fill values of a matrix with structure identical to A with
        # selected entries from the difference of tensor products:
        #   b_Ax (X) H - AH (X) x
        # NOTE: this row-wise manipulation can be much faster in C++ or Cython
        A_colInd = ctx.sparse_structure.col_ind
        A_rowPtr = ctx.sparse_structure.row_ptr
        batch_size = grad_output.shape[0]
        A_grad = torch.empty(
            size=(batch_size, len(A_colInd)), **targs
        )  # return value, A's grad
        for r in range(len(A_rowPtr) - 1):
            start, end = A_rowPtr[r], A_rowPtr[r + 1]
            columns = A_colInd[start:end]  # col indices, for this row
            A_grad[:, start:end] = (
                b_Ax[:, r].unsqueeze(1) * H[:, columns]
                - AH[:, r].unsqueeze(1) * ctx.x[:, columns]
            )

        # apply correction if there is a multiplicative damping
        if ctx.damping_alpha_beta is not None and ctx.damping_alpha_beta[0] > 0.0:
            alpha = ctx.damping_alpha_beta[0]
            alpha2Hx = (alpha * 2.0) * H * ctx.x  # componentwise product
            A_grad -= ctx.A_val * alpha2Hx[:, ctx.A_colInd.type(torch.long)]

        return A_grad, AH, None, None, None, None, None, None
